// Delaunay.cpp: Triangulation
// 2007-06-28
//
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <vector>
using namespace std;

#include "vUtil.h"
#include "igeom.h"			//    
#include "pvectors.h"		//    

#include "Delaunay.h"		//  


//   (  )
int		addTriangle(pvector &tr, pnt &v1, pnt &v2, pnt &v3, int &ind); //    ,     .
int		triangleSq2(pnt &v1, pnt &v2, pnt &v3);				// 2*  
void	getFirstEdge(pvector &p, int &index0, int &index1); //    &p     (  )
int		MostShortEdge(pvector &edges);						//     
double	TwoT( pnt &a, pnt &b, pnt &c );						//   <a,b,c>   2*t, 
void	join_v( pnt &a, pnt &b, pvector &p, pvector &res );	//   
double  Cangl(pnt &p1, pnt &p2, pnt &p3);					//   (p1,p2,p3)
int		triangleSq2(pnt &v1, pnt &v2, pnt &v3);				// triangleSq((0,0),(1,0),(0,1))->1
void	add2(pvector &a, pvector &tr, pnt &v1, pnt &v2);	//  (v1,v2)  a,        tr

//void	drawTr(pvector &tr, char *bmp);						//    

//   

/////////////////////////////////////////////////////////////
bool	SameP(pnt &v1, pnt &v2)							//    ?
{
	if (v1.x == v2.x && v1.y == v2.y) return true;
	return false;
}
bool	SameE(pnt &v1, pnt &v2, pnt &v3, pnt &v4)			//    (v1,v2)  (v3,v4) ?
{
	if (SameP(v1, v3) && SameP(v2, v4)) return true;
	if (SameP(v1, v4) && SameP(v2, v3)) return true;
	return false;
}
bool SameT(pnt &v1, pnt &v2, pnt &v3, pnt &w1, pnt &w2, pnt &w3) {	//   
	if (SameP(v1, w1)) return SameE(v2, v3, w2, w3);
	if (SameP(v1, w2)) return SameE(v2, v3, w1, w3);
	if (SameP(v1, w3)) return SameE(v2, v3, w1, w2);
	return false;
}
/////////////////////////////////////////////////////////////
int  edgeInTri(pvector &tr, pnt &v1, pnt &v2)				//    [v1,v2]  - 
{
	int k = 0;
	for (int i = 0; i<tr.size(); i += 3)
		for (int j = 0; j < 3; j++, k++) {
			if (SameP(tr[k], v1)) {
				if (j == 0) {
					if (SameP(tr[k + 1], v2) || SameP(tr[k + 2], v2)) return i;
				}
				if (j == 1) {
					if (SameP(tr[k - 1], v2) || SameP(tr[k + 1], v2)) return i;
				}
				if (j == 2) {
					if (SameP(tr[k - 2], v2) || SameP(tr[k - 1], v2)) return i;
				}
			}
		}
	return -1;
}
/////////////////////////////////////////////////////////////////////////////
//    ,     .
// ind -   .
//  0,      1,   .
int addTriangle(pvector &tr, pnt &v1, pnt &v2, pnt &v3, int &ind)
{
	ind = -1;
	for (int i = 0; i<tr.size(); i += 3)
		if (SameT(tr[i], tr[i + 1], tr[i + 2], v1, v2, v3)) return 0;
	//     . .			
	ind = tr.size();
	tr.add(v1);
	tr.add(v2);
	tr.add(v3);
	return 1;
}
/////////////////////////////////////////////////////////////
//    ,      !!!
void	add2(pvector &a, pvector &tr, pnt &v1, pnt &v2)	//  (v1,v2)  a,        tr
{
	for (int i = 0; i<a.size(); i += 2) {
		if(SameE(v1, v2, a[i], a[i+1]))
		{
			a.del(i, i + 1);
			return;
		}
	}
	if (edgeInTri(tr, v1, v2)>=0) return;					//   .   
	a.add(v1);
	a.add(v2);
}
/////////////////////////////////////////////////////////////
// 2*  
// triangleSq((0,0),(1,0),(0,1))->1
int triangleSq2(pnt &v1, pnt &v2, pnt &v3)
{
	return (v2.x - v1.x)*(v3.y - v1.y) - (v2.y - v1.y)*(v3.x - v1.x);
}
/////////////////////////////////////////////////////////////
void	swapPoints(pnt &p1, pnt &p2)						//  
{
	swapInt(p1.x, p2.x);
	swapInt(p1.y, p2.y);
}
/////////////////////////////////////////////////////////////
double  Cangl(pnt &p1, pnt &p2, pnt &p3)					//   (p1,p2,p3)
{
	//return CanglE(A[0] - B[0], A[1] - B[1], C[0] - B[0], C[1] - B[1]);
	return CanglE(p1.x-p2.x, p1.y-p2.y, p3.x-p2.x, p3.y-p2.y);
}
/////////////////////////////////////////////////////////////
int    EdgeLen2(pvector &edges, int i) //   i- 
{
	return sumSq(edges.getX(i)- edges.getX(i+1), edges.getY(i) - edges.getY(i + 1));
}
/////////////////////////////////////////////////////////////
//    &p     (  )
void getFirstEdge(pvector &p, int &index0, int &index1)
{
	//   a0 -   
	int min_x = 999999999;
	int    i, min_i;
	for (i = 0; i<p.size(); i++) {
		int x = p.getX(i);
		if (x<min_x) { min_x = x; min_i = i; }
	}
	index0 = min_i;
	//          -
	//       (dy/dx - )
	double k_min = 2e99;
	int min_diff = 999999999;
	for (i = 0; i<p.size(); i++) {
		if (i == index0) continue;
		int dx = p.getX(i) - p.getX(index0);			//   (index0)   ,  dx>=0
		int dy = p.getY(i) - p.getY(index0);			// dy - 
		double kA = dx==0? -2e99: dy/ (double)dx;
		if (kA<k_min) {
			k_min = kA;
			min_diff = abs(dx) + abs(dy);
			min_i = i;
			continue;
		}
		//    ,    :
		if (kA == k_min) {
			int diff1 = abs(dx) + abs(dy);
			if (diff1<min_diff) { min_diff = diff1; min_i = i; }
		}
	}
	index1 = min_i;
}

/////////////////////////////////////////////////////////////////////////////
//     
int MostShortEdge(pvector &edges)
{
	int len = 2147483647; // INT_MAX;
	int imin = -1;
	for (int i = 0; i < edges.size(); i += 2)
		if (EdgeLen2(edges, i) < len) {
			imin = i; len = EdgeLen2(edges, i);
		}
	return imin;
}
/////////////////////////////////////////////////////////////
//   <a,b,c>   2*t, 
//        
//    .
//
double TwoT(pnt &a, pnt &b, pnt &c)
{
	double d = double(c.y - b.y)*(b.x - a.x) - double(c.x - b.x)*(b.y - a.y);
	if (d == 0) return 1e99;
	return (double(c.x - b.x)*(a.x - c.x) + double(c.y - b.y)*(a.y - c.y)) / d;
}

/////////////////////////////////////////////////////////////
//    <a,b>   p[i]  p  , 
// 1)    <a,b,p[i]>  0
// 2) ,    <a,b,p[i]>    
//       ( p[i]),   (1)
//    .  (2)   ,   TwoT 
//         .
// ,      .
//      res
//
void join_v(pnt &a, pnt &b, pvector &p, pvector &res) //   
{
	double min_2t = 1e99;
	res.clear();
	for (int i = 0; i<p.size(); i++) {
		pnt c;
		c.x = p.getX(i);
		c.y = p.getY(i);

		if (triangleSq2(a, b, c) <= 0) continue;
		double TwoT_ = -TwoT(a, b, c);
		if (TwoT_ == min_2t) res.add(c); //         
		if (TwoT_<min_2t) { min_2t = TwoT_; res.clear(); res.add(c); } //   ,  
	}
}


/////////////////////////////////////////////////////////////
//  .
//    &p
//   tr    :
// tr[0], tr[1], tr[2], tr[3], tr[4], tr[5], tr[6], tr[7], tr[8], ...
//  <tr[0], tr[1], tr[2]> -   0- ,
//  <tr[3], tr[4], tr[5]> -   1- ,
// ....
void Delaunay(pvector &p, pvector &tr)
{
	tr.clear();
	if (p.size()<3) return;
	int i0, i1; //    
	int i, j;
	getFirstEdge(p, i0, i1);
	pvector edges;
	pvector adj; //   
	pnt v1, v2, v3;
	edges.add(p[i0]);     edges.add(p[i1]);
	//   
	while (edges.size()>0) { //    
		//edges.save("edges.txt", "edges", 2);
		//tr.save("tr.txt", "tri", 3);
		//drawTr(tr, "a.bmp");
		i = MostShortEdge(edges);
		v1 = edges[i];
		v2 = edges[i+1];
		edges.del(i, i + 1);
		// (v1,v2) -     edges.  .
		join_v(v1, v2, p, adj);
		// adj -     (v1,v2)
		if (adj.size() == 0) continue;	//     -   .
		if (adj.size() == 1) //     
		{
			v3 = adj[0];
			if (triangleSq2(v1, v2, v3) <= 0) vError("Triangulation: 2*square <= 0");
			//     
			add2(edges, tr, v1, v3);			// (v1,v3)
			add2(edges, tr, v3, v2);			// (v3,v2)
			addTriangle(tr, v1, v2, v3, i);
			continue;
		}
		//  ,    
		//        <v2, v1, adj[i]>
		//         <v2, v1, adj[i]>
		for (i = 1; i<adj.size(); i++)
		{
			double a0 = Cangl(v2, v1, adj[i - 1]);
			double a1 = Cangl(v2, v1, adj[i]);
			if (a0<a1)
			{
				//swapPoints(*adj.getPoint(i), *adj.getPoint(i - 1));
				swapPoints(adj[i-1], adj[i]);
				if (i>1) i -= 2;
			}

		}
		//adj.save("adj.txt", "adj", 1);
		//   <v2, v1, adj[i]> 
		//  
		// v1, v2,     adj[0],
		// v1, adj[0], adj[1],
		// ...
		// v1, adj[k-2], adj[k-1],
		add2(edges, tr, adj[0], v2);					// (adj[i], v2)
		addTriangle(tr, v1, v2, adj[0], j);

		for (i = 1; i<adj.size(); i++)
		{
			add2(edges, tr, adj[i], adj[i-1]);			// (adj[i], adj[i-1])
			addTriangle(tr, v1, adj[i], adj[i-1], j);
		}
		add2(edges, tr, v1, adj[adj.size() - 1]);		// (v1, adj[last])
	}

	//       ,     >0
	for (i = 0; i < tr.size(); i += 3)
		if (triangleSq2(tr[i], tr[i + 1], tr[i + 2]) < 0)
			//swapPoints(*tr.getPoint(i), *tr.getPoint(i + 1));
			swapPoints(tr[i], tr[i+1]);

}
//////////////////////////////////////////////////
//     
void saveTriangles(pvector &triangles, char *filename, char*msg) {
	FILE *f = fopen(filename, "w");
	fprintf(f, "%s, #%d\n", msg, triangles.size() / 3);
	for (int i = 0; i<triangles.size(); i += 3)
		fprintf(f, "%4d) (%4d,%4d), (%4d,%4d), (%4d,%4d)\n", i / 3,
			triangles.getX(i), triangles.getY(i),
			triangles.getX(i + 1), triangles.getY(i + 1),
			triangles.getX(i + 2), triangles.getY(i + 2));
	fclose(f);
}



///////////////////////////////////////////////////////////
// Triangles
////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////



/////////////////////////////////////////////////////////////////////////////
//      
//      tr.
//    k <tr.getPoint(3*k), tr.getPoint(3*k+1),tr.getPoint(3*k+2)>
//     
// r=1 -   tr.getPoint(3*k+1),tr.getPoint(3*k+2)
// r=2 -   tr.getPoint(3*k  ),tr.getPoint(3*k+2)
// r=3 -   tr.getPoint(3*k  ),tr.getPoint(3*k+1)
//   ,  -1
// !!!!!        3 !!!!!!!!!1
int  adjoinTriangle( pvector &tr, int k, int r )
{
	pnt *v1, *v2,*v3;
	if( r<1 || r>3 ) vError( "adjoinTriangle: r=%d", r );
	if(r==1)
	{
		v1=&tr.get(3*k  );
		v2=&tr.get(3*k+1);
		v3=&tr.get(3*k+2);
	}
	if(r==2) 
	{ 
		v1 = &tr.get(3 * k + 1);
		v2 = &tr.get(3 * k);
		v3 = &tr.get(3*k+2);
	}
	if(r==3) 
	{ 
		v1 = &tr.get(3 * k + 2);
		v2 = &tr.get(3 * k + 1);
		v3 = &tr.get(3*k  );
	}
	//   <*v1,*v2,*v3>     <*v2,*v3>
	for( int i=0; i<tr.size(); i+= 3 )
	{
		if( SameE(*v2,*v3, tr.get(i), tr.get(i+1)) &&
		   !SameP(*v1, tr.get(i+2))) return i/3;
			
		if( SameE(*v2,*v3, tr.get(i+1),tr.get(i+2)) &&
		   !SameP(*v1, tr.get(i))) return i/3;
			
		if( SameE(*v2,*v3, tr.get(i+2),tr.get(i)) &&
		   !SameP(*v1, tr.get(i+1))) return i/3;
			
	}
	return -1;
}
////////////////////////////////////////////////////////////////////////
// <*t, *(t+1), *(t+2)> -> <*(t+1), *(t+2), *t>
void rotateTrangle(pnt *t)
{
	pnt tmp;
	tmp.x = t->x;			tmp.y = t->y;
	t->x = (t + 1)->x;		t->y = (t + 1)->y;
	(t + 1)->x = (t + 2)->x;(t + 1)->y = (t + 2)->y;
	(t + 2)->x = tmp.x;		(t + 2)->y = tmp.y;
}
//////////////////////////////////////////////////////////////////////////////
//   it[0..19]   ,    p
//   ind[0..19] -   p     (0|1|2)
//   
//   :
// <tr[it[0]],tr[it[0]+1],tr[it[0]+2]>, p = tr[it[0]+ind[0]]
// <tr[it[1]],tr[it[1]+1],tr[it[1]+2]>, p = tr[it[1]+ind[1]]
// ...
//
// :
//     
//		frame.triangles.getPoint( it[k  ] )
//		frame.triangles.getPoint( it[k+1] )
//		frame.triangles.getPoint( it[k+2] )
//   frame.points.getPoint(ip)    
//		frame.triangles.getPoint( it[k+ind[k]] )
int getListOfTrianglesInPoint(pvector &tr, pnt &p, int *it, int *ind)
{
	int cnt = 0;
	for (int t = 0; t<tr.size(); t += 3) // t -  
	{
		bool b = false;
		int n = 0;
		if (tr.get(t)==p)
		{
			b = true;
		}
		else if (tr.get(t + 1)==p)
		{
			b = true;
			n = 1;
		}
		else if (tr.get(t + 2)==p)
		{
			b = true;
			n = 2;
		}

		if (b)	//    
		{
			if (cnt == MAX_TRIANGLE) vError("getListOfTrianglesInPoint: too many triangles"); // return -1;  
			ind[cnt] = n;
			it[cnt++] = t;
		}
	}
	return cnt;
} // getListOfTrianglesInPoint




/*

/////////////////////////////////////////////////////////////////////////////
//   <(x1,y1),(x2,y2),(x3,y3)>
//      b1(x,y), b2(x,y), b3(x,y):
// b1(x,y) = b1x*x + b1y*y + b1_;
// b2(x,y) = b2x*x + b2y*y + b2_;
// b3(x,y) = b3x*x + b3y*y + b3_;
void baricentr( double x1,  double y1,   double x2,  double y2,   double x3,  double y3,  
				double &b1x, double &b1y, double &b1_,
				double &b2x, double &b2y, double &b2_,
				double &b3x, double &b3y, double &b3_ )
{
	double den =  x3*y1 - x3*y2 + x1*y2 - x1*y3 + x2*y3 - x2*y1;
	if( den==0 ) vError("baricentr: triangle is degenerated");	//  
	b1x  = (y2-y3)/den;
	b1y  = (x3-x2)/den;
	b1_  = (x2*y3-x3*y2)/den;

	b2x  = (y3-y1)/den;
	b2y  = (x1-x3)/den;
	b2_  = (x3*y1-x1*y3)/den;

	b3x  = (y1-y2)/den;
	b3y  = (x2-x1)/den;
	b3_  = (x1*y2-x2*y1)/den;

}	


//////////////////////////////////////////////////////////////////////////////
//    edges   [a,b]
int findEdge( pvector &edges, pnt &a, pnt &b )
{
	for( int i=0; i<edges.getSize(); i+= 2)
	{
		if( edges.getPoint(i  )->equals(a) &&
			edges.getPoint(i+1)->equals(b) ) return i;
		if( edges.getPoint(i  )->equals(b) &&
			edges.getPoint(i+1)->equals(a) ) return i;
	}
	return -1;
}


//////////////////////////////////////////////////////////////////////////////
char* trangleToString( pnt *t, char *buf ) // <*t, *(t+1), *(t+2)> ->  
{
	sprintf( buf,
		" tr<(%4d,%4d)(%4d,%4d)(%4d,%4d)>, area=%8d; ",
		t[0].x, t[0].y, t[1].x, t[1].y, t[2].x, t[2].y, triangleSq2( t[0], t[1], t[2]) );
	return buf;
}
*/
//////////////////////////////////////////////////////////////////////////////
