// dMatrix.h: matrix of ldouble
// 2017-06-06

//
#ifdef DMATRIX
#error Secondary include
#endif
#define DMATRIX


class dvect {	//    
private:
	int mx;			// dimension
	ldouble *v;		// vector
public:
	// constructors/destructor
	dvect(int mx_=0);
	dvect( dvect &a );
	~dvect();
	// memory, sizes
	ldouble *elem() { return v; }
	void	release();		// release memory
	void	clear();		//   
	//    ,    ,  ,   
	void	setSize		(int mx_);	//  
	void	setSizeClear(int mx_);	//    
	void	copyFrom(dvect &a);
	void	copyFrom(int N, ldouble *x);
	// tests
	void	test();
    void	testSameSize(dvect &a);
	void	testIndexes(int x);		// 0<=x<max
	void	normaIndexes(int &x);	//    0<=x<max
	// get/put
	int		Mx() { return mx; }					// 
	int		size() { return mx; }				// 

	ldouble &operator() (int x);				//     (  )
	ldouble &operator[] (int x);				//     (  )

	// in/out
	void	saveText(const char *fname, const char *msg = "", const char *format = " %12.5f");
	void	appendText(const char *fname, const char *msg = "", const char *format = " %12.5f");
	void	loadText(const char *fname);
	void	inc(int x, ldouble value);			// w[x] += value

	void	 set( ldouble value );				//    
	ldouble  minV( int &x );					//   , x -  
	ldouble  maxV(int &x);						//   , x -  
	ldouble  sum();								//  
	// algebra
	void	add ( dvect &v1, ldouble z);		// this += z*v1
	void	add ( ldouble *v1, ldouble z);		// this += z*v1
	void	add(ldouble z);						//   += z
	void	mult( ldouble z);					//     v
	void	killZero();							//    
    void    normalize();    		            //     =1
	ldouble	length();    				        // 
	ldouble	length2();    				        //  
	ldouble	dist(dvect &vv);    		        //   
    ldouble	dist2(dvect &vv);    		        //    
    ldouble	scProd(dvect &vv);    		        //  
	//
	void	setA(int mx_, ...);					//    
	void	setRandom(ldouble mi, ldouble ma);	//     
	void	append(ldouble z);					//     
	// sort
	void	sort();								//   

};
//=============================================================================
// dmatr
//=============================================================================
class dmatr {
private:
	int mx, my;		// dimension
	ldouble *v;		// matrix
	//ldouble lambda;  // A -> T(A)*A + lambda*E
	ldouble eps;		//  0

	bool	isMultipl(ldouble b, ldouble eps);	//     b    eps)
public:
	// constructors/destructor
	dmatr(int mx_ = 0, int my_ = 0);
	dmatr(dmatr &a);
	~dmatr();
	// memory, sizes
	void	release();		// release memory
	void	clear();		//   
	//    ,    ,  ,   
	void	setSize(int mx_, int my_);	//  
	void	setSizeClear(int mx_, int my_);	//    
	void	copyFrom(dmatr &a);
	ldouble	setEps();		//      eps ( 0)
	ldouble  getEps() { return eps; }
	// tests
	void	test();
	void	testSameSize(dmatr &a);
	void	testIndexes(int x, int y);	// 0<=x,y<max
	void	normaIndexes(int &x, int &y);	//     0<=x,y<max
											// get/put
	int		Mx() { return mx; }
	int		My() { return my; }
	int		getSize() { return mx*my; }
	ldouble &operator() (int x, int y);				//     (  )
	ldouble *getElement(int x, int y);
	ldouble	get(int x, int y);
	ldouble	getLast(int y);						//   
	void	put(int x, int y, ldouble value);
	void	safePut(int x, int y, ldouble value);
	void	inc(int x, int y, ldouble value);	// w[x,y] += value
	void	set(ldouble value);					//     
	void	setRandom(ldouble mi, ldouble ma);	//     
	void    getLine(int y, dvect &v);			// v:= line(y)
	void    getColumn(int x, dvect &v);			// v:= column(x)
	// in/out
	void	saveText(const char *fname, const char *msg = "", const char *format = " %12.5f");
	void	saveText2(const char *fname, const char *msg = "");
	void	saveTextI(const char *fname, const char *msg = "", const char *format = " (%4d,%4d=>%12.5f)"); //   
	void	appendText(const char *fname, const char *msg = "", const char *format = " %12.5f");
	void	loadText0(const char *fname);
	//void	loadText1(const char *fname);
	void	loadText(const char *fname);
	// algebra
	ldouble	minimal(int &x_, int &y_);			// minimal value
	ldouble	minimal();							// minimal value
	ldouble	maximal(int &x_, int &y_);			// maximal value
	ldouble	maximal();							// maximal value
	ldouble	maxCol(int x, int &y_);				// maximal value in column
	ldouble	minCol(int x, int &y_);				// minimal value in column
	ldouble	maxLine(int &x_, int y);			// maximal value in line
	ldouble	minLine(int &x_, int y);			// minimal value in line
	ldouble	sumLine(int y);						// summa values in line
	ldouble	sumCol(int x);						// summa values in column

	void	copyLine(dmatr &a, int y, int ya);	// y- (this) := ya- ()
	void	copyLine(dvect &a, int y);			// y- (this) := 
	void	copyCol (dmatr &a, int x, int xa);	// x- (this) := xa- ()
	void	copyBlock(dmatr &a, int x0, int y0, int dx, int dy);// this := (a)  dx*dy
	
	void	add(dmatr &a);						// this += a
	void	add(ldouble z);						//   += z
	void	mult(ldouble z);					//     z
	void	mult(dvect &dst, dvect &src);		// dst := this*src; src.size    this->mx
	void	mult(dmatr &a, dmatr &b);			// this := a*b
	void	multL(dvect &dst, dvect &src);		// dst := src*this; src.size    this->my
	ldouble	bilin(dvect &L, dvect &R);			// retrun sum a[i,j]*L[i]*R[j]
	void	multLine(int y, ldouble z);			//    v
	void	multCol(int x, ldouble z);			//    v
	void	addMult(int y, int y1, ldouble z);	//      y1,   z
	void	addL(int y, dvect &r, ldouble z);	//      r,   z
	void	swapLine(int y, int y1);			//   y, y1
	void	swapCol(int x, int x1);				//   x,x1
	void	killZero();							//    
	void	killCol(int x, int y);				//    [x,y]     
	ldouble	maxInColDown(int x, int &y);		//  w[x,y] -       x   y
	int		triangulate();						//    .  
	void	addSq(ldouble *u);					// v[k,l]+=u[k]*u[l]
	void    normalize();    		            //     =1
	void    mTranspose(dmatr &a);				// this := transpose(a)*a
	void    mTransp(dmatr &a);					// this := transpose(a)*a,   my=mx-1

	ldouble	lineLength(int y);					//   y- 
	ldouble	columnLength(int x);				//   x- 

	ldouble	isOrthonormal();			  		//     
	ldouble isSymmetric();			  			// return difference for symmtricity
	void    setFourierBasis(int N);				//    f   -  N
	bool    NormaLine(int y);					//    y   1. false -    0
	bool    NormaCol(int x);					//    x   1. false -    0
	ldouble ScProd(int y1, int y2);				//   
	ldouble ScProdC(int x1, int x2);			//   
	ldouble ScProd(int y, dvect &r);			//   y-    r
	void    GramShmidt();						//        
	void    transpose(dmatr &a);				// this :=transpose(a)
	void    transpose();						//  " "  

	//  
	int		solve(dvect &sol);					// sol :=,  
	ldouble det();								//   ,  
	bool	revers(dmatr &a0);					// this := revers(a0)

	//    evec     (this)!
	void    EigenQL(dvect &eval, dmatr &evec);	// eigenvalues (increased) and eigenvectors of symmetric matrix, ev:=eigenvalues(this), v:=eigenvectors(this)

	int		solve2(dvect &sol);					// sol := ,  

};
