// dfunc2.cpp  functions of two variables f(x,y)
// by Sergei Khashin 2017-06-06
#include <stdlib.h>
#include <math.h>

#include "vUtil.h"			//   
#include "dMatrix.h"		//  
#include "dfunc1.h"		    // functions of one variable f(x)
#include "dfunc.h"			// functions of n variable f(dvect)

//----------------------------------------------------------------------------
// ,  min[i] <= v[i] <= max[i]
//   alpha , 
// ,  min[i] <= v[i] + alpha*g[i] <= max[i]
double maxAlpha(dvect &v, dvect &min, dvect &max, dvect &g)
{
	int N = v.size();
	double al = dbl_MAX, a2;
	for (int i = 0; i < N; i++) {
		double gi = g[i];
		if (gi == 0) continue;
		if (gi > 0) {
			a2 = (max[i] - v[i]) / gi;		// a2 >0
			if (a2 < al) al = a2;
		}
		else {
			a2 = (min[i] - v[i]) / gi;		// a2>0 !
			if (a2 < al) al = a2;
		}
	}
	return al;
}
//-----------------------------------------------------------------------------
class MathFunc1V : public MathFunc1
{
public:
	~MathFunc1V() {};
	MathFunc *fN;
	dvect *v_;
	dvect *g_;
	dvect u;		//  
	double operator () (double t);
};

//-----------------------------------------------------------------------------
double MathFunc1V::operator () (double t)
{
	u.copyFrom(*v_);
	u.add(*g_, t);
	return (*fN)(u);
}
//-----------------------------------------------------------------------------
void MathFunc::setProt(char *prot_)
{
	prot = prot_;
}
//-----------------------------------------------------------------------------
// g :=     v, eps -  
void MathFunc::gradient(dvect &v, dvect &g, double eps)
{
	int N = v.size();
	for (int i = 0; i < N; i++) {
		double t = v[i];					//   
		v[i] += eps;
		double f1 = this->operator()(v);
		v[i] = t - eps;
		double f2 = this->operator()(v);
		v[i] = t;							//   
		g[i] = 0.5*(f1 - f2) / eps;
	}
}
//-----------------------------------------------------------------------------
double MathFunc::minGrad(dvect &v, dvect &min, dvect &max, double epsX, double epsF)
{
	int N = v.size();
	//  min <= v <= max
	for (int i = 0; i < N; i++) {
		double t = v[i];
		if (t <min[i] || t>max[i]) vError("MathFunc::minGrad: i=%d, min = %10.5f, v = %10.5f, max=%10.5f", i, min[i], t, max[i]);
	}
	dvect grad(N); //  

	//    
	MathFunc1V *fvv = new MathFunc1V;
	fvv->v_ = &v;
	fvv->g_ = &grad;
	fvv->u.setSize(N);
	fvv->fN = this;
	fv = fvv;
	// , 

	double fv1 = this->operator()(v);	//   
	//v.appendText("v.txt", toString("\n--------------------\n     val=%10.5f; ", fv1));
	double fv0;							//   
	int iStep = 0;
	do {
		//v.appendText("prot.txt", toString("iStep=%4d, val=%10.5f; ", iStep, fv1));
		saveStepInfo(iStep++, v);
		gradient(v, grad, epsX);		// grad -  ()  
		double gradLen = grad.length();
		if (gradLen < 1e-9) break;		//   , 
		//    f( v - alpha*grad), alpha = [0..aMax].  aMax 
		grad.mult(-1);
		double aMax = maxAlpha(v, min, max, grad);	//   alpha
		double al = fmin(0, aMax, *fv, epsX);		// al,    
		v.add(grad, al);
		fv0 = fv1;
		fv1 = this->operator()(v);					//   
		if( prot)
			v.appendText(prot, toString("val=%10.5f; ", fv1));
		if (fv1 >= fv0) {							//  (),  
			v.add(grad, -al);
			fvv->~MathFunc1V();
			return fv0;
		}
		if (fv1 > fv0 - epsF) break;				//   
		if (al*gradLen < epsX) break;				//   
	} while (1);
	fvv->~MathFunc1V();
	return fv1;
}
//-----------------------------------------------------------------------------
//   
double MathFunc::minCoord(dvect &v, dvect &min, dvect &max, double epsX, double epsF)
{
	int N = v.size();
	double h = 0.5;			//  ,     max[i]-min[i]
	double maxDiff = 0;		//    max-min
	for (int i = 0; i < N; i++)
		if (max[i] - min[i] > maxDiff) maxDiff = max[i] - min[i];
	epsX /= maxDiff;		//  ,   h<epsX
	double fCurr = this->operator()(v);	//   
	double f1;							//   
	do {
		h *= 0.5;						//      h
		bool change;					//        
		do {
			change = false;
			for (int i = 0; i < N; i++) {	//   
				double hi = h*(max[i] - min[i]);
				double vi = v[i];
				v[i] += hi;
				if (v[i] <= max[i]) {
					f1 = this->operator()(v);//   
					if (f1 < fCurr) {
						fCurr = f1;
						if (prot)
							v.appendText(prot, toString("val=%10.5f; ", f1));
						change = true;
						continue;			//   
					}
				}
				v[i] = vi - hi;
				if (v[i] >= min[i]) {
					f1 = this->operator()(v);//   
					if (f1 < fCurr) {
						fCurr = f1;
						if( prot)
							v.appendText(prot, toString("val=%10.5f; ", f1));
						change = true;
						continue;			//   
					}
				}
				v[i] = vi;
			}// for (int i = 0; i < N; i++) {	//   
		} while (change);						//    
	} while (h > epsX);
	return fCurr;
}
//-----------------------------------------------------------------------------
// voff     2*N*(N-1)
static void makeVectors(int N, dvect *voff)	//     N
{
	if (N == 2) {							//  ,   6 
		const double be = sqrt(3.0)/2;
		voff[0][0] =  1;   voff[0][1] =   0;		// 0
		voff[1][0] =  0.5; voff[1][1] =  be;		// 1
		voff[2][0] = -0.5; voff[2][1] =  be;		// 2
		voff[3][0] = -1;   voff[3][1] =   0;		// 3
		voff[4][0] = -0.5; voff[4][1] = -be;		// 4
		voff[5][0] =  0.5; voff[5][1] = -be;		// 5
		return;
	}
	const double al = 1 / sqrt(2.0);
	int k = 0;
	for( int i=1; i<N; i++)
		for (int j = 0; j < i; j++) {
			voff[k].setSizeClear(N);
			voff[k][i] = al;
			voff[k][j] = al;
			k++;

			voff[k].setSizeClear(N);
			voff[k][i] = al;
			voff[k][j] =-al;
			k++;

			voff[k].setSizeClear(N);
			voff[k][i] =-al;
			voff[k][j] = al;
			k++;

			voff[k].setSizeClear(N);
			voff[k][i] =-al;
			voff[k][j] =-al;
			k++;
		}
}

double MathFunc::minCoor2(dvect &v, dvect &min, dvect &max, double epsX, double epsF) //    N<=6
{
	int N = v.size();
	if (N > 6) return minCoord(v, min, max, epsX, epsF);	//   > 6,    
	//   
	double maxDiff = 0;		//    max-min
	for (int i = 0; i < N; i++)
		if (max[i] - min[i] > maxDiff) maxDiff = max[i] - min[i];
	epsX /= maxDiff;		//  ,   h<epsX
	double fCurr = this->operator()(v);	//   
	double f1;							//   
	double h = 0.05*maxDiff;			//  
	int nv = 2*N*(N-1);					//  
	if (N == 2) nv = 6;
	dvect *voff = new dvect[nv];
	makeVectors(N, voff);
	dvect w(N);								//    
	do {
		h *= 0.5;							//      h
		bool change;						//        
		do {
			change = false;
			for (int i = 0; i < nv; i++) {	//    
				w.copyFrom(v);
				w.add(voff[i], h);			// w -   
				f1 = this->operator()(w);	//   
				if (f1 < fCurr-epsF) {		//  ,     
					fCurr = f1;
					v.copyFrom(w);
					if (prot)
						v.appendText(prot, toString("val=%10.5f; i=%3d, h=%10.2g", f1, i, h));
					do {
						w.copyFrom(v);
						w.add(voff[i], h);			// w -   
						f1 = this->operator()(w);	//   
						if (f1 >= fCurr) break;		//  ,  
						//   
						fCurr = f1;
						v.copyFrom(w);
						if (prot)
							v.appendText(prot, toString("val=%10.5f; i=%3d, h=%10.2g", f1, i, h));
					} while (1);
					change = true;
				}
			}// for (int i = 0; i < nv; i++) {	//    
		} while (change);						//    
	} while (h > epsX);
	delete[]voff;
	return fCurr;
}

//-----------------------------------------------------------------------------


