// iMatrix.cpp: matrix with integer coefficients
// 2006-11-06
#ifdef _MSC_VER
  #pragma warning(disable : 4996)
#endif

#ifdef __GNUC__
#pragma GCC diagnostic ignored "-Wwrite-strings"
#pragma GCC diagnostic ignored "-Wformat"
#pragma GCC diagnostic ignored "-Wsign-compare"
#endif
//
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <math.h>
#include <limits.h>

#include "vUtil.h"		    //   
//#include "igeom.h"			//    ,     edgeIterator
#include "iMatrix.h"		//  



//-----------------------------------------------------------------------------
edgeIterator::edgeIterator(int x0_, int y0_, int x1_, int y1_)
{
	x0 = x0_; y0 = y0_; x1 = x1_; y1 = y1_;
	deltaX = abs(x1 - x0);
	deltaY = abs(y1 - y0);
	signX = x0 < x1 ? 1 : -1;
	signY = y0 < y1 ? 1 : -1;
	error = deltaX - deltaY;
}
edgeIterator::edgeIterator(int *P0, int *P1)
{
	x0 = P0[0]; y0 = P0[1]; x1 = P1[0]; y1 = P1[1];
	deltaX = abs(x1 - x0);
	deltaY = abs(y1 - y0);
	signX = x0 < x1 ? 1 : -1;
	signY = y0 < y1 ? 1 : -1;
	error = deltaX - deltaY;
}
bool edgeIterator::next()		// ,   
{
	if (x0 == x1 && y0 == y1) return true;
	const int error2 = error * 2;
	if (error2 > -deltaY) {
		error -= deltaY;
		x0 += signX;
	}
	if (error2 < deltaX) {
		error += deltaX;
		y0 += signY;
	}
	return false;
}
//-----------------------------------------------------------------------------
static int 	iDivide_( int a, int b ) // round(a/b)
{
	int sign=1;
	if( a<0 ) {a=-a; sign=-1; }
	int d = (a+b/2)/b;
	return sign>0 ? d : -d;
}
static int  bias=0;
static int 	iDivide_bias( int a, int b ) // round(a/b) with bias
{
	int sign=1;
	if( a<0 ) {a=-a; sign=-1; }
	int v = a+b/2;
	int d = v/b;
	if( v%b==0 ) {d-=bias; bias=1-bias; }
	return sign>0 ? d : -d;
}

#define vMatrixTextIndexes	//    

////////////////////////////////////////////////////
// constructors/destructors
////////////////////////////////////////////////////
//   mx*my.
matr::matr(int mx_, int my_)
{
	v=NULL; mx=0; my=0; this->setSize(mx_,my_);
}

matr::matr(matr &a)
{
	vError("matr::matr(matr &a): copy constructor illegal");
}

matr::~matr() { this->release(); }

void	matr::release()
{
	if( mx<=0 || my<=0 ) return;
	if (v != NULL) delete []v;
	v = NULL;
	mx=my=0;
}

//////////////////////////////////////////////////////////////////////
void	matr::clear()		//   
{
	int size = mx*my;
	if(size>0)
	{
		//int *p=v; for( int i=0; i<size; i++ ) *p++=0;
		memset(v,0,size*sizeof(*v));
	}
}
//////////////////////////////////////////////////////////////////////
void	matr::setSize( int mx_, int my_) //    ,   
{
	if( mx_<0 || my_<0) vError( "matr::setSize: mx=%d, my=%d", mx, my );
	if( mx==mx_ && my==my_ ) return; //    ,   
	if( v != NULL )
	{
		delete [] v;
		v=NULL;
	}
	mx = mx_;
	my = my_;
	int size = mx*my;
	if(size>0)
	{
		v = new int[size];
		//int *p=v; for( int i=0; i<size; i++ ) *p++=0;
		memset(v,0,size*sizeof(*v));
	}
	else { v = NULL; }
}

//////////////////////////////////////////////////////////////////////
void	matr::setSizeClear( int mx_, int my_) //    
{
	if( mx_<0 || my_<0) vError( "matr::setSize: mx=%d, my=%d", mx, my );
	if( mx==mx_ && my==my_ ) //    , 
	{
		clear();
		return;
	}
	if( v != NULL )
	{
		delete [] v;
		v=NULL;
	}
	mx = mx_;
	my = my_;
	int size = mx*my;
	if(size>0)
	{
		v = new int[size];
		//int *p=v; for( int i=0; i<size; i++ ) *p++=0;
		memset(v,0,size*sizeof(*v));
	}
	else { v = NULL; }
}

//////////////////////////////////////////////////////////////////////
void	matr::copyFrom(matr &a)
{
	a.test();
	this->setSize( a.Mx(), a.My() );
	//for( int i=0; i<a.getSize(); i++ ) v[i] = a.v[i];
	memcpy(v, a.v, a.getSize()*sizeof(*v));
}
//////////////////////////////////////////////////////////////////////
void matr::test()
{
	//if( v==NULL) vError( "matr test: v==NULL" );
	if( mx<=0 || my<=0 ) vError( "matr test: mx=%d, my=%d", mx, my );
	if((mx>0 || my>0) && (v==NULL) ) vError( "matr: sizes>0, but v==NULL" );
	if( mx>100000 || my>100000 ) vError( "matr::test mx=%d, my=%d", mx, my );
}

void	matr::testSameSize( matr &a  )
{
	test();	a.test();
	if( mx!=a.mx || my!=a.my )
		vError( "matr::testSameSize: mx=%d, a.mx=%d, my=%d, a.my=%d",
		                                 mx,    a.mx,    my,    a.my );
}

void	matr::testIndexes(int x, int y) // 0<=x,y<max
{
	if(x<0 || x>=mx || y<0 || y>=my) vError("matr::testIndexes x=%d, y=%d",x,y);
}

bool   matr::isIndexes(int x, int y)		// ?0<=x,y<max
{
	return  x>=0 && x < mx && y>=0 && y < my;
}

void	matr::normaIndexes(int &x, int &y)	//     0<=x,y<max
{
	if( x<  0) x=0;
	if( x>=mx) x=mx-1;
	if( y<  0) y=0;
	if( y>=my) y=my-1;
}

int *matr::getElement(int x, int y)
{
#ifdef	vMatrixTextIndexes	//  
	this->testIndexes(x,y);
#endif
	return v + y*mx + x;
}


void	matr::copyLine  ( int yDst, int ySrc ) //  
{
	copyBuf( getElement(0,yDst), 1, getElement(0,ySrc), 1, mx );
}
void	matr::copyColumn( int xDst, int xSrc ) //  
{
	copyBuf( getElement(xDst,0), mx, getElement(xSrc,0), mx, my );
}

//    (x0,y0, x1,y1)  v
void	matr::fillOuter( int x0, int y0, int x1, int y1, int v)
{
	int x,y;
	for(y=0 ; y<y0; y++ ) this->fillLine(y,v);
	for(y=y1+1; y<my; y++ ) this->fillLine(y,v);
	for(x=0 ; x<x0; x++ ) this->fillColumn(x,v);
	for(x=x1+1; x<mx; x++ ) this->fillColumn(x,v);
}

///////////////////////////////////////////////////////////
// get/put

int matr::get( int x, int y )
{
#ifdef	vMatrixTextIndexes	//  
	this->testIndexes(x,y);
#endif
	return v[y*mx + x];
}

void	matr::put( int x, int y, int value )
{
#ifdef	vMatrixTextIndexes	//  
	this->testIndexes(x,y);
#endif
	v[y*mx + x] = value;
}

void	matr::inc( int x, int y, int value )		//     value
{
#ifdef	vMatrixTextIndexes	//  
	this->testIndexes(x,y);
#endif
	v[y*mx + x] += value;
}

int &matr::operator () (int x, int y)
{
#ifdef	vMatrixTextIndexes	//  
	this->testIndexes(x, y);
#endif
	return v[y*mx + x];
}


int		matr::safeGet( int x, int y )			//     ,   
{
	normaIndexes(x,y);
	return v[y*mx + x];
}

void	matr::safePut( int x, int y, int value ) //  ,       
{
	if(x<0 || x>=mx || y<0 || y>=my ) return;
	//put(x,y,value);
	v[y*mx + x] = value;
}

void	matr::set( int value )
{
	int *p=v;
	//for( int i=mx*my; i>0; i-- ) *p++=value;
	for( int i=mx*my; i--; ) *p++=value;
}

/////////////////////////////////////////////////////////////////////////
// private
void	matr::fillLine  ( int y0, int v )  //    v
{
	for( int x=0; x<mx; x++ ) put(x,y0,v);
}
/////////////////////////////////////////////////////////////////////////
// private
void	matr::fillColumn( int x0, int v )  //    v
{
	for( int y=0; y<my; y++ ) put(x0,y,v);
}

///////////////////////////////////////////////////////////////
ldouble	matr::getInterp( ldouble x, ldouble y, int k ) // universal interpolation
{
	if( k==0 ) return this->safeGet( round(x), round(y) );
	if( k==1 ) return this->getInterp1(x,y);
			   return this->getInterp3(x,y);
}

///////////////////////////////////////////////////////////////
// bilinear interpolation for v(x,y)
ldouble	matr::getInterp1( ldouble x, ldouble y ) // bilinear get( x, y )
{
	int x0, y0;		//    x,y
	ldouble dx, dy;	//    x,y

	if     (x<    0) {x0=0   ; dx=0; }
	else if(x>=mx-1) {x0=mx-1; dx=0; }
	else     { x0 = (int)x; dx=x-x0; }

	if     (y<    0) {y0=0   ; dy=0; }
	else if(y>=my-1) {y0=my-1; dy=0; }
	else     { y0 = (int)y; dy=y-y0; }

	if( dx==0 && dy==0 ) return get(x0,y0);
	if( x0==(mx-1)|| y0==(my-1) ) return get(x0,y0);

	int f00 = this->get(x0,y0);
	int f10 = this->get(x0+1,y0);
	int f01 = this->get(x0,y0+1);
	int f11 = this->get(x0+1,y0+1);

	ldouble val = f00*(1-dx)*(1-dy) +
		         f10*dx    *(1-dy) +
			     f01*(1-dx)*dy     +
			     f11*dx*dy;
	return val;
}

///////////////////////////////////////////////////////////////
//Description: cubic interpolation (by 4 points)
//Arguments: given 4 defined values
//      fm=f(-1), f0=f(0), f1=f(1), f2=f(2);
// and interpolating point 'x'
//Return value: interpolated value for interpolating point
static ldouble  Cubic( ldouble fm, ldouble f0, ldouble f1, ldouble f2, ldouble x ){
  return  f0
          +(f1-1./6*f2-1./3*fm-1./2*f0)*x
          +(1./2*f1-f0+1./2*fm)*x*x
          +(1./6*f2-1./2*f1+1./2*f0-1./6*fm)*x*x*x;
};
///////////////////////////////////////////////////////////////
// bicubic interpolation for v(x,y)
ldouble	matr::getInterp3( ldouble x, ldouble y ) // bicubic get( x, y )
{
	int x0, y0;		//    x,y
	ldouble dx, dy;	//    x,y

	if     (x<    0) {x0=0   ; dx=0; }
	else if(x>=mx-1) {x0=mx-1; dx=0; }
	else     { x0 = (int)x; dx=x-x0; }

	if     (y<    0) {y0=0   ; dy=0; }
	else if(y>=my-1) {y0=my-1; dy=0; }
	else     { y0 = (int)y; dy=y-y0; }

	if( dx==0 && dy==0 ) return get(x0,y0);
	//if( x0==(mx-1)|| y0==(my-1) ) return get(x0,y0);

    ldouble fm = Cubic( safeGet(x0-1, y0-1),
	                   safeGet(x0  , y0-1),
                       safeGet(x0+1, y0-1),
                       safeGet(x0+2, y0-1), dx );
    ldouble f0 = Cubic( safeGet(x0-1, y0  ),
                       safeGet(x0  , y0  ),
                       safeGet(x0+1, y0  ),
                       safeGet(x0+2, y0  ), dx );
    ldouble f1 = Cubic( safeGet(x0-1, y0+1),
                       safeGet(x0  , y0+1),
                       safeGet(x0+1, y0+1),
                       safeGet(x0+2, y0+1), dx );
    ldouble f2 = Cubic( safeGet(x0-1, y0+2),
                       safeGet(x0  , y0+2),
                       safeGet(x0+1, y0+2),
                       safeGet(x0+2, y0+2), dx );
	return Cubic( fm, f0, f1, f2, dy );
} // ldouble	matr::getInterp3( ldouble x, ldouble y ) // bicubic get( x, y )
////////////////////////////////////////////////////////////////////////////
// in/out
void	matr::saveText( const char *fname, const char *msg, const char *format )
{
	this->test();
	FILE *f = fopen( fname, "w" );
	if(!f) vError( "saveText: file not open" );
	if(msg)
		fprintf( f, "%4d %4d %s\n", mx, my, msg );
	else
		fprintf( f, "%4d %4d\n", mx, my );
	for( int y=0; y<my; y++ )
	{
		for( int x=0; x<mx; x++ )
		{
			fprintf( f, format, this->get(x,y) );
			//fprintf( f, " (%4d,%4d=>%6d)", x, y, this->get(x,y) );
		}
		fprintf( f, "\n");
	}
	fclose(f);
}

void	matr::appendText(const char *fname, const char *msg, const char *format)
{
	this->test();
	FILE *f = aopen(fname);
	if (msg)
		fprintf(f, "%4d %4d %s\n", mx, my, msg);
	else
		fprintf(f, "%4d %4d\n", mx, my);
	for (int y = 0; y<my; y++)
	{
		for (int x = 0; x<mx; x++)
		{
			fprintf(f, format, this->get(x, y));
			//fprintf( f, " (%4d,%4d=>%6d)", x, y, this->get(x,y) );
		}
		fprintf(f, "\n");
	}
	fclose(f);
}

void	matr::saveTextI(const char *fname, const char *msg, const char *format ) //   
{
	this->test();
	FILE *f = fopen( fname, "w" );
	if(!f) vError( "saveText: file not open" );
	if(msg)
		fprintf( f, "%4d %4d %s\n", mx, my, msg );
	else
		fprintf( f, "%4d %4d\n", mx, my );
	for( int y=0; y<my; y++ )
	{
		for( int x=0; x<mx; x++ )
		{
			//fprintf( f, format, this->get(x,y) );
			fprintf( f, format, x, y, this->get(x,y) );
		}
		fprintf( f, "\n");
	}
	fclose(f);
}

void	matr::saveTextB(const char *fname, int x0, int y0, int dx, int dy, const char *msg, const char *format) //   
{
	this->test();
	FILE *f = fopen(fname, "w");
	if (!f) vError("saveText: file not open");
	if (msg)
		fprintf(f, "%4d %4d %s\n", mx, my, msg);
	else
		fprintf(f, "%4d %4d\n", mx, my);
	for (int y = y0; y<y0+dy; y++)
	{
		for (int x = x0; x<x0+dx; x++)
		{
			//fprintf( f, format, this->get(x,y) );
			fprintf(f, format, x, y, this->get(x, y));
		}
		fprintf(f, "\n");
	}
	fclose(f);
}


void	matr::loadText ( const char *fname )
{
	this->release();
	FILE *f = fopen( fname, "r" );
	if(!f) vError( "matr::loadText: file %s not open", fname );
	char s[255]; int mx_, my_;
	fscanf( f, "%d%d", &mx_, &my_);
	if( mx_<=0 || my_ <= 0 ) vError("matr::loadText: mx=%d, my=%d", mx, my);
	fgets( s, 255, f );
	//if( mx<my ) { mx += my; my = mx-my; mx -= my; }
	this->setSize(mx_,my_);
	int t;
	for( int y=0; y<my; y++ )
	for( int x=0; x<mx; x++ )
	{
		fscanf( f, "%d", &t );
		this->put(x,y,t);
	}
	fclose(f);
}


////////////////////////////////////////////////////////////////////////
void	matr::copyBlock( matr &a, int x0, int y0, int dx, int dy )
{
	if (x0 < 0) { dx += x0;	x0 = 0; if (dx <= 0) return; }
	if (x0 >= a.Mx()) return;
	if (x0 + dx >= a.Mx()) dx = a.Mx() - 1 - x0;

	if (y0 < 0) { dy += y0;	y0 = 0;  if (dy <= 0) return;}
	if (y0 >= a.My()) return;
	if (y0 + dy >= a.My()) dy = a.My() - 1 - y0;

	this->setSize(dx, dy);
	for( int y=0; y<dy; y++ )
	for( int x=0; x<dx; x++ )
		this->put(x,y, a.get(x+x0,y+y0) );
}

////////////////////////////////////////////////////////////////////////
void	matr::copyBlock( matr &a, int xSrc, int ySrc, int xDst, int yDst, int dx, int dy )
{
	if( dx<1 || dy<1 ) vError("matr::copyBlock: dx=%d; dy=%d", dx, dy);
	int aMx = a.Mx(), aMy = a.My();
	int mx = this->Mx(), my = this->My();

	if( xSrc<0 || xSrc>=aMx) vError("matr::copyBlock: xSrc=%d", xSrc);
	if( ySrc<0 || ySrc>=aMy) vError("matr::copyBlock: ySrc=%d", ySrc);

	if(xSrc+dx>aMx ) dx = aMx-xSrc;
	if(ySrc+dy>aMy ) dy = aMy-ySrc;

	if(xDst<0 || yDst<0 ) vError("matr::copyBlock: xDst=%d; yDst=%d", xDst, yDst);

	int mx1=mx, my1=my;
	if(xDst+dx>mx ) mx1=xDst+dx;
	if(yDst+dy>my ) my1=yDst+dy;

	if(mx1>mx || my1>my ) { this->setSize(mx1,my1); mx=mx1; my=my1; }

	for( int ddy=0; ddy<dy; ddy++ )
	for( int ddx=0; ddx<dx; ddx++ )
		this->put(xDst+ddx,yDst+ddy, a.get(xSrc+ddx,ySrc+ddy) );
}

void	matr::fillBlock( int v, int x0, int y0, int dx, int dy )
{
	if( x0<0 ) x0=0; 	if( dx>=mx-x0 ) dx = mx-x0;
	if( y0<0 ) y0=0; 	if( dy>=my-y0 ) dy = my-y0;

	for( int y=y0; y<y0+dy; y++ )
	for( int x=x0; x<x0+dx; x++ )
		this->put(x,y, v );
}

void	matr::fillCircle( int v,int xr, int yr, int r )
{
	int x0=xr-r, x1=xr+r, y0=yr-r, y1=yr+r;
	if( x0<0 ) x0=0; 	if( x1>=mx ) x1 = mx-1;
	if( y0<0 ) y0=0; 	if( y1>=my ) y1 = my-1;
	int r2 = r*r;
	for( int y=y0; y<=y1; y++ )
	for( int x=x0; x<=x1; x++ )
		if( (y-yr)*(y-yr) + (x-xr)*(x-xr) <= r2 )
			this->put(x,y, v );
}
/////////////////////////////////////////////////////////////////////
void	matr::swap( matr &a )	// this <-> a
{
	int tmx = mx, tmy = my;
	int *tv = v;

	mx = a.mx;
	my = a.my;
	v  = a.v;

	a.mx = tmx;
	a.my = tmy;
	a.v  = tv;
}

/////////////////////////////////////////////////////////////////////
// min, max, count, change
int		matr::minimal(int &x_, int &y_) // minimal value
{
	int vi, imin, min_=INT_MAX;
	for( int i=mx*my-1; i>=0; i-- )
		if( (vi=v[i])<min_ ) { imin=i; min_=vi; }
	y_ = imin/mx; x_ = imin%mx;
	return min_;
}

int		matr::minimal()				// minimal value
{
	int vi, min_=INT_MAX;
	for( int i=mx*my-1; i>=0; i-- )
		if( (vi=v[i])<min_ ) { min_=vi; }
	return min_;
}

int		matr::maximal(int &x_, int &y_) // maximal value
{
	int vi, imax, max_=INT_MIN;
	for( int i=mx*my-1; i>=0; i-- )
		if( (vi=v[i])>max_ ) { imax=i; max_=vi; }
	y_ = imax/mx; x_ = imax%mx;
	return max_;
}

int		matr::maximal()				// maximal value
{
	int vi, max_=INT_MIN;
	for( int i=mx*my-1; i>=0; i-- )
		if( (vi=v[i])>max_ ) { max_=vi; }
	return max_;
}

int	matr::minLin(int &x_, int y_)			// minimal value in line y, (x,y) - point
{
	int W = 0x7fffffff, t;
	for (int x = 0; x < mx; x++) {
		t = get(x, y_);
		if (t < W) {
			W = t;
			x_ = x;
		}
	}
	return W;
}
int	matr::maxLin(int &x_, int y_)			// maximal value in line y, (x,y) - point
{
	int W = 0x80000000, t;
	for (int x = 0; x < mx; x++) {
		t = get(x, y_);
		if (t > W) {
			W = t;
			x_ = x;
		}
	}
	return W;
}
int	matr::minCol(int x_, int &y_)			// minimal value in column x, (x,y) - point
{
	int W = 0x7fffffff, t;
	for (int y = 0; y < my; y++) {
		t = get(x_, y);
		if (t < W) {
			W = t;
			y_ = y;
		}
	}
	return W;
}
int	matr::maxCol(int x_, int &y_)			// maximal value in column x, (x,y) - point
{
	int W = 0x80000000, t;
	for (int y = 0; y < my; y++) {
		t = get(x_, y);
		if (t > W) {
			W = t;
			y_ = y;
		}
	}
	return W;
}


ldouble  matr::mid()								// mid value
{
	ldouble su=0;
	for( int i=mx*my-1; i>=0; i-- ) su += v[i];
	return su/(mx*my);
}

//////////////////////////////////////////////////////////////////////
//   [(x0,y0),(x1,y1)]  c
void	matr::drawEdge(int x0, int y0, int x1, int y1, int c) {
	edgeIterator ei(x0, y0, x1, y1);
	do {
		x0 = ei.X();
		y0 = ei.Y();
		if (c>0) this->safePut(x0, y0, c);
		else     this->safePut(x0, y0, (128 + this->safeGet(x0, y0)) % 256);
	}
	while (!ei.next());
}

ldouble  matr::midEdge(int x0, int y0, int x1, int y1) // mid value on the edge (x0,y0)-(x1,y1)
{
	int cnt = 0, sum = 0;
	edgeIterator ei(x0, y0, x1, y1);
	do {
		sum += this->get(ei.X(), ei.Y());
		cnt++;
	} while (!ei.next());

	return ((ldouble)sum) / cnt;
}

ldouble  matr::midEdg2(int x0, int y0, int x1, int y1) // mid*sqrt(len)
{
	int cnt = 0, sum = 0;
	edgeIterator ei(x0, y0, x1, y1);
	do {
		sum += this->get(ei.X(), ei.Y());
		cnt++;
	} while (!ei.next());
	return sum / pow(cnt, 0.9);
}

int	 matr::minEdge(int x0, int y0, int x1, int y1, int &x2, int &y2) // min value on the edge (x0,y0)-(x1,y1)
{
	int min = 0x7fffffff, m1;
	x2 = x0;
	y2 = y0;
	edgeIterator ei(x0, y0, x1, y1);
	do {
		if ((m1 = this->get(ei.X(), ei.Y()))<min)
		{
			min = m1;
			x2 = ei.X();
			y2 = ei.Y();
		}
	} while (!ei.next());
	return min;
}
int	 matr::maxEdge(int x0, int y0, int x1, int y1, int &x2, int &y2) // max value on the edge (x0,y0)-(x1,y1)
{
	int max = 0x80000000, m1;
	x2 = x0;
	y2 = y0;
	edgeIterator ei(x0, y0, x1, y1);
	do {
		if ((m1 = this->get(ei.X(), ei.Y()))<max)
		{
			max = m1;
			x2 = ei.X();
			y2 = ei.Y();
		}
	} while (!ei.next());
	return max;
}

//-----------------------------------------------------------------------------
ldouble matr::midBox(int x0, int y0, int x1, int y1) // mid value on the box (x0,y0)-(x1,y1)
{
	normaIndexes(x0, y0);
	normaIndexes(x1, y1);
	int xMin= INT_MAX, xMax = INT_MIN, yMin = INT_MAX, yMax = INT_MIN;
	corrMinMax(x0, xMin, xMax);
	corrMinMax(y0, yMin, yMax);
	int cnt = 0, sum = 0;
	for (int y = yMin; y <= yMax; y++)
	for (int x = xMin; x <= xMax; x++) {
		sum += this->get(x, y);
		cnt++;
	}
	return ((ldouble)sum) / cnt;
}
//-----------------------------------------------------------------------------
int		matr::countVal( int val )	  //  ,  v
{
	int count=0;
	for( int i=mx*my-1; i>=0; i-- )
		if( v[i]==val ) count++;
	return count;
}

int		matr::countValI( int val0, int val1 )	//   x: val0<=x<val1
{
	int count=0;
	for( int i=mx*my-1; i>=0; i-- )
	{
		int vi = v[i];
		if( val0<=vi && vi<val1 ) count++;
	}
	return count;
}


int		matr::changeVal(int v1, int v2) //  v1  v2.   
{
	int count=0;
	for( int i=mx*my-1; i>=0; i-- )
		if( v[i]==v1 ) { count++; v[i]=v2; }
	return count;
}

int		matr::changeVal2(int v1, int v2, int v3)	//     [v1..v2]  v3.   
{
	int count=0;
	int vi;
	for( int i=mx*my-1; i>=0; i-- )
	{
		vi = v[i];
		if( v1<=vi && vi<=v2 ) { count++; v[i]=v3; }
	}
	return count;
}

// ,  min   min, max - 
void	matr::clip( int min_, int max_ )
{
	//int count=0;
	for( int i=mx*my-1; i>=0; i-- )
	{
		//int vi = v[i];
		if( v[i]<min_ ) { v[i]=min_; }
		if( v[i]>max_ ) { v[i]=max_; }
	}
}

void matr::normalize( int k )				//       [0..k]
{
	this->addV( -this->minimal() );			//   
	int mmax = this->maximal();
	if (mmax == 0) return;
	this->mult( INT_MAX/2/mmax);			//    [0..MAX_INT/2]
	this->div ( (this->maximal()+k-1)/k );	//    [0..k]
}
void matr::Mod( int k )						//     mod k (0..k-1)
{
	for( int i=mx*my-1; i>=0; i-- )
	{
		v[i] = mod( v[i], k);
	}
}

//////////////////////////////////////////////////////////////
// algebra
//////////////////////////////////////////////////////////////
void matr::mult( int k) 					// this *= k
{	test();
	for( int i=0; i<getSize(); i++ ) v[i] *= k;
}
void matr::mult( matr &a) 				// this(x,y) *= a(x,y) 
{	this->testSameSize(a);
	for( int i=0; i<getSize(); i++ ) v[i] *= a.v[i];
}

void matr::div ( int k) 					// this /= k
{	test();
	if( k==0 ) vError("matr::div, k=0");
	for( int i=0; i<getSize(); i++ ) v[i] /= k;
}
void matr::add( matr &a) 				// this += a
{	this->testSameSize(a);
	for( int i=0; i<getSize(); i++ ) v[i] += a.v[i];
}
void matr::add( matr &a, int k)			// this += a*k
{	this->testSameSize(a);
	for( int i=0; i<getSize(); i++ ) v[i] += a.v[i]*k;
}
void matr::addV( int r) 					// this[*,*] += r
{	test();
	for( int i=0; i<getSize(); i++ ) v[i] += r;
}

void matr::sub( matr &a) 				// this -= a
{	this->testSameSize(a);
	for( int i=0; i<getSize(); i++ ) v[i] -= a.v[i];
}
void matr::addMult( matr &a, int k)// this += a*k
{	this->testSameSize(a);
	for( int i=0; i<getSize(); i++ ) v[i] += k*a.v[i];
}
void matr:: mask(matr &a)						//  a(x,y)!=0,  this(x,y):=a(x,y)
{
	this->testSameSize(a);
	int t;
	for (int i = 0; i < getSize(); i++)	if( (t= a.v[i])!=0) v[i] = t;
}
void matr::mask(matr &a, int x0, int y0)		//  a(x,y)!=0,  this(x0+x,y0+y):=a(x,y)
{
	for (int y = 0; y<a.My(); y++)
		for (int x = 0; x < a.Mx(); x++) {
			int t = a(x, y);
			if (t != 0) safePut(x0+x, y0+y, t);
		}
}
void matr::addB(matr &a)						// this += a 
{
	this->testSameSize(a);
	for (int i = 0; i<getSize(); i++) v[i] = byteAdd(v[i], a.v[i]);
}
int	matr::sumRow( int y)	//   y- 
{
	int su=0;
	for( int x=0; x<mx; x++ ) su += get(x,y);
	return su;
}

int	matr::sumColumn( int x)	//   x- 
{
	int su=0;
	for( int y=0; y<my; y++ ) su += get(x,y);
	return su;
}

ldouble matr::CSprite(matr &spr, int x, int y)	//      
{
	if (x < 0) x = 0; if (x >= mx - 1) return 0;
	if (y < 0) y = 0; if (y >= my - 1) return 0;

	int dx = spr.Mx(); if (x + dx > mx) dx = mx - x;
	int dy = spr.My(); if (y + dy > my) dy = my - y;

	ldouble ra = 0, rs = 0, ras = 0;
	for (int y1 = y; y1<y + dy; y1++)
		for (int x1 = x; x1 < x + dx; x1++) {
			int qa = get(x1, y1);
			int qs = spr.get(x1 - x, y1 - y);
			ra += qa*qa;
			rs += qs*qs;
			ras += qa*qs;
		}
	ldouble L = ra*rs;
	if (L == 0) return 0;
	return ras / sqrt(L);
}
////////////////////////////////////////////////////////
// partial derivative
//   2,    

////////////////////////////////////////////////////////////
int		matr::dx( int x, int y)				// 2*da/dx  
{
	this->normaIndexes(x, y);
	if( x==0 )			return 2*(get(1,y) - get(0,y));
	else if(x==(mx-1) )	return 2*(get(x,y) - get(x-1,y));
	else				return get(x+1,y)  - get(x-1,y);
}
int		matr::dy( int x, int y)				// 2*da/dy  
{
	this->normaIndexes(x, y);
	if( y==0 )			return 2*(get(x,1) - get(x,0));
	else if(y==(my-1) )	return 2*(get(x,y) - get(x,y-1));
	else				return get(x,y+1)  - get(x,y-1);
}
int	matr::dxx(int x, int y)					// d2a/dxdx in the point
{
	this->normaIndexes(x, y);
	if (x == 0)	x++;
	else if (x == mx - 1) x--;
	return get(x - 1, y) - 2 * get(x, y) + get(x + 1, y);
}
int	matr::dxy(int x, int y)					// d2a/dxdy in the point
{
	this->normaIndexes(x, y);
	if (x == 0)	x++;
	else if (x == mx - 1) x--;
	if (y == 0)	y++;
	else if (y == my - 1) y--;
	return (2 + get(x + 1, y + 1) - get(x + 1, y - 1) - get(x - 1, y + 1) + get(x - 1, y - 1)) / 4;
}
int	matr::dyy(int x, int y)					// d2a/dydy in the point
{
	this->normaIndexes(x, y);
	if (y == 0)	y++;
	else if (y == my - 1) y--;
	return get(x, y - 1) - 2 * get(x, y) + get(x, y + 1);
}
double matr::Rinvar(int x, int y)				// R-  
{
	double fx = dx(x, y);
	double fy = dy(x, y);
	double fxx = dxx(x, y);
	double fxy = dxy(x, y);
	double fyy = dyy(x, y);
	return 0.25*(fx*fx*fxx + 2 * fx*fy*fxy + fy*fy*fyy);
}
int	   matr::Rinvars(int x, int y)				// (-1|0|1)  R-  
{
	double R = Rinvar(x, y);
	return R < 0 ? -1 : R>0 ? 1 : 0;
}

ldouble matr::l2Grad( int x, int y)	 //    
{
	int Dx = dx(x,y);
	int Dy = dy(x,y);
	ldouble len = sqrt((ldouble)(Dx*Dx+Dy*Dy))/2;
	return len;
}
void matr::l2Grad( matr &a)		//     
{
	int mx = a.Mx(), my = a.My();
	this->setSize(mx,my);
	int x,y;
	for( y=0; y<my; y++ )
	for( x=0; x<mx; x++ )
	{
		ldouble Dx = a.dx(x,y);
		ldouble Dy = a.dy(x,y);
		this->put(x,y, round(sqrt(Dx*Dx+Dy*Dy)/2) );
	}
}

//////////////////////////////////////////////////////////////////////
int matr::mGrad(int x, int y, int h)		// morphological gradient in a point (max-min amoung (2*h+1)*(2*h+1) neigbour points)
{
	int minV=INT_MAX, maxV=INT_MIN, v;
	for( int y1=y-h; y1<=y+h; y1++ )
	for( int x1=x-h; x1<=x+h; x1++ )
	{
		v = safeGet(x1,y1);
		if( v<minV ) minV=v;
		if( v>maxV ) maxV=v;
	}
	return maxV-minV;
}

void matr::mGrad ( matr &a, int h) // this:= morphological gradient in each point(a) (max-min amoung (2*h+1)*(2*h+1) neigbour points)
{
	setSize(a.Mx(), a.My());
	int x,y;
	for( y=0; y<my; y++ )
	for( x=0; x<mx; x++ )
	{
		put(x,y, a.mGrad(x,y, h) );
	}
}
///////////////////////////////////////////////////////////////////////////////////////
int matr::variat(int x, int y, int h)	//         (2*h+1)*(2*h+1)
{
	int sum=0, v0 = get(x,y);
	for (int y1 = y - h; y1 <= y + h; y1++)
		for (int x1 = x - h; x1 <= x + h; x1++)
		{
			int v = safeGet(x1, y1);
			sum += abs(v - v0);
		}
	return sum;

}
void matr::variat(matr &a, int h)			// this:=         (2*h+1)*(2*h+1)
{
	setSize(a.Mx(), a.My());
	int x, y;
	for (y = 0; y<my; y++)
		for (x = 0; x<mx; x++)
		{
			put(x, y, a.variat(x, y, h));
		}
}

double matr::midVar()							//    
{
	double S = 0;
	int cnt = 0;
	for(int y = 0; y<my; y++)
	for(int x = 0; x<mx; x++)
	{
		int v0 = get(x, y);
		if (y > 0) {
			S += iabs(get(x, y - 1) - v0);
			cnt++;
		}
		if (x > 0) {
			S += iabs(get(x - 1, y) - v0);
			cnt++;
		}
	}
	return S / cnt;
}
///////////////////////////////////////////////////////////////////////////////////////
// dst =    src  bufSize    r
//    
void smoothLine0( int *dst, int *src, int bufSize, int r ) // dst=smooth(src, radius=r);
{
	//    .
	//  ( ),  busSize>r
	//   sum -     [i-r,i+r]  2*r+1.
	//     (r+1)*src[0] + src[1] + .. + src[r]
	int i, rr = 2*r+1;
	int v0 = src[0], vlast = src[bufSize-1];
	int sum = (r+1)*src[0];
	for( i=1; i<=r; i++ ) sum += src[i];
	//   sum 
	for( i=0; i<bufSize; i++ )
	{
		dst[i] = iDivide_(sum, rr);
		sum -= (i<r           ? v0   : src[i-r  ] );  //  sum
		sum += (i>=bufSize-r-2? vlast: src[i+r+1] );  //  sum
	}
}
///////////////////////////////////////////////////////////////////////////////////////
// dst =    src  bufSize    r
//  
void smoothLine1( int *dst, int *src, int bufSize, int r ) // dst=smooth(src, radius=r);
{
	int sum, count, i,i0, i1, j, offs;
	for( i=0; i<bufSize; i++ )
	{
		i0 = i-r; if(i0<0) i0=0;
		i1 = i+r; if(i1>=bufSize) i1 = bufSize-1;
		sum=0; count=0;
		for( j=i0; j<=i1; j++ )
		{
			offs = r+1-abs(j-i);
			sum   += src[j]*offs;
			count += offs;
		}
		dst[i] = iDivide_(sum, count);
	}
}
///////////////////////////////////////////////////////////////////////////////////////
// dst =    src  bufSize    r
//  
/*
   R(t,r) = C(r) * exp( -t^2/r^2/2 );

*/
static const int Gauss_M = 256;		//    
static ldouble Gauss_C[3*Gauss_M];	//  
static int Gauss_R=0;				//   
//
static void smoothLine2_calc( int &r )	//    Gauss_C
{

	if( r==Gauss_R) return;				//   ,    
	if( r<1 ) { Gauss_C[0]=r=0; return; }
	if( r>=Gauss_M ) r = Gauss_M-1;
	Gauss_R=r;
	ldouble s=1, rr=0.5/(r*r);
	Gauss_C[0]=1;
	int x;
	for( x=1; x<=3*r; x++ ) { Gauss_C[x]=exp(-rr*x*x); s+=2*Gauss_C[x]; }
	s = 1/s;
	for( x=0; x<=3*r; x++ ) Gauss_C[x]*=s;

	//s = Gauss_C[0];
	//for( x=1; x<=3*r; x++ ) s+=2*Gauss_C[x];
	//s += 1;
}

static ldouble smoothLine2_sum( int i0, int i1 )	//    (Gauss_C)  i0  i1
// ,  i0<=0<=i1
{
	if( i0<=-3*Gauss_R && i1>=3*Gauss_R ) return 1;
	ldouble su=Gauss_C[0];
	int i;
	for( i=i0; i<0;   i++ ) su += Gauss_C[-i];
	for( i=1;  i<=i1; i++ ) su += Gauss_C[i];
	return 1./su;
}


void smoothLine2( int *dst, int *src, int bufSize, int r ) // dst=smooth(src, radius=r);
{
	if( r<1 ) return;
	smoothLine2_calc(r);
	int r3 = 3*r;
	ldouble s;
	int i,j, j0, j1;
	for( i=0; i<bufSize; i++ )
	{
		j0 = i-r3; if( j0<0 ) j0=0;
		j1 = i+r3; if( j1>=bufSize ) j1=bufSize-1;
		for( j=j0, s=0; j<=j1; j++ ) s += Gauss_C[abs(j-i)]*src[j];
		s *= smoothLine2_sum( j0-i, j1-i);
		dst[i]=round(s);
	}
}
///////////////////////////////////////////////////////////////////////////////////////
// dst =    src  bufSize    r
// kind = 0 ( ) -    
// kind = 1                -  
// kind = 2                -   ()
void smoothLine( int *dst, int *src, int bufSize, int r, int kind ) // dst=smooth(src, radius=r);
{
	if( kind==0 ) { smoothLine0( dst, src, bufSize, r); return; }
	if( kind==1 ) { smoothLine1( dst, src, bufSize, r); return; }
	if( kind==2 ) { smoothLine2( dst, src, bufSize, r); return; }
}
/////////////////////////////////////////////////////////////////////////
// kind = 0                -    
// kind = 1 ( ) -  
// kind = 2                -   ()
void	matr::smooth( int r, int kind )	//    r
{
	if( r<1 ) return;		//   <1     
	if( r>=mx-2 ) r=mx-2;	//    
	if( r>=my-2 ) r=my-2;	// 
	int bufSize = imax2(mx,my);
	int *buf1 = new int[bufSize];
	int *buf2 = new int[bufSize];
	//this->saveText("t:\\0.txt");

	int mult_coeff;								//  
	int maximal_ =  this->maximal();
	if( maximal_==0 ) return;
	mult_coeff = (1<<24)/(r*r*maximal_);	//  
	if( mult_coeff > 2) mult(mult_coeff);	//    [0..2^24]

	//this->mult(256);	//    256   
	//this->saveText("t:\\1.txt", NULL, "%9d");
	int x,y;
	//   :
	for( y=0; y<my; y++ )
	{
		copyBuf( buf1, 1, this->getElement(0,y), 1, mx );	//    buf1
		smoothLine( buf2, buf1, mx, r, kind );
		copyBuf( this->getElement(0,y), 1, buf2, 1, mx );	//    buf2  
	}

	//this->saveText("t:\\2.txt", NULL, "%9d");
	//   :
	for( x=0; x<mx; x++ )
	{
		copyBuf( buf1, 1, this->getElement(x,0), mx, my );	//    buf1
		smoothLine( buf2, buf1, my, r, kind );
		copyBuf( this->getElement(x,0), mx, buf2, 1, my );	//    buf2  
	}
	//this->saveText("t:\\3.txt", NULL, "%9d");
	if( mult_coeff > 2) div(mult_coeff);	//    
	//this->saveText("t:\\4.txt", NULL, "%9d");
	delete []buf1;
	delete []buf2;
}

//////////////////////////////////////////////////////////////////////////////////////////
void matr::decrease( matr &a, int k )	// this :=    k (a)
{
	this->setSize( a.Mx()/k, a.My()/k );
	for( int y=0; y<my; y++ )
	for( int x=0; x<mx; x++ )
	{
		int s=0;
		for( int dy=0; dy<k; dy++ )
		for( int dx=0; dx<k; dx++ ) s+=a.get(k*x+dx, k*y+dy);
		this->put(x,y, k&1? iDivide_(s, k*k): iDivide_bias(s, k*k));
	}
}
//////////////////////////////////////////////////////////////////////////////////////////
void matr::resize( matr &a, int mx1, int my1)	// this := resize(a)  (mx,my)
{
    int x,y;
    this->setSize(mx1,my1);
    ldouble x0, y0, cox, coy;
    cox = a.mx; cox /= mx1;
    coy = a.my; coy /= my1;

    for( y=0; y<my; y++ )
    {
        y0 = y*coy;
        for( x=0; x<mx; x++ ) {
            x0 = x*cox;
            this->put(x,y, round(a.getInterp3(x0,y0)));
        }
    }
}
//////////////////////////////////////////////////////////////////////////////////////////
void matr::diffX( matr &a )					// this := 2*da/dx
{
	this->setSize( a.Mx(), a.My() );
	for( int y=0; y<my; y++ )
	for( int x=0; x<mx; x++ )
		this->put(x,y, a.dx(x,y));
}
//////////////////////////////////////////////////////////////////////////////////////////
void matr::diffY( matr &a )					// this := 2*da/dy
{
	this->setSize( a.Mx(), a.My() );
	for( int y=0; y<my; y++ )
	for( int x=0; x<mx; x++ )
		this->put(x,y, a.dy(x,y));
}
//////////////////////////////////////////////////////////////////////////////////////////
void matr::diffXX(matr &a)					// this := d^2a/dxdx
{
	this->setSize(a.Mx(), a.My());
	for (int y = 0; y < my; y++) {
		for (int x = 1; x < mx-1; x++)
			this->put(x, y, a.get(x-1, y)-2*a.get(x,y)+a.get(x+1,y));
		this->put(0, y, this->get(1, y));
		this->put(mx-1, y, this->get(mx-2, y));
	}

}
void matr::diffXY(matr &a)					// this := d^2a/dxdy
{
	this->setSize(a.Mx(), a.My());
	for (int y = 1; y < my - 1; y++) {
		for (int x = 1; x < mx-1; x++)
			this->put(x, y, (2+a.get(x+1, y + 1) -  a.get(x+1, y-1) - a.get(x - 1, y + 1) + a.get(x - 1, y - 1))/4);
	}
	for (int x = 0; x < mx; x++) {
		this->put(x, 0, this->get(x, 1));
		this->put(x, my - 1, this->get(x, my - 2));
	}
	for (int y = 1; y < my-1; y++) {
		this->put(0, y, this->get(1, y));
		this->put(mx-1, y, this->get(mx-2, y));
	}
}
void matr::diffYY(matr &a)					// this := d^2a/dxdx
{
	this->setSize(a.Mx(), a.My());
	for (int y = 1; y < my-1; y++) {
		for (int x = 0; x < mx; x++)
			this->put(x, y, a.get(x, y-1) - 2 * a.get(x, y) + a.get(x, y + 1));
	}
	for (int x = 0; x < mx; x++) {
		this->put(x, 0, this->get(x, 1));
		this->put(x, my - 1,  this->get(x, my - 2));
	}
}
/////////////////////////////////////////////////////////////////
void matr::minNeib( matr &a, int r ) // this[x,y] :=   a    r
{
	a.test();
	int r2 = r*r;
	setSize( a.Mx(), a.My() );
	int minV, currV;
	for( int y=0; y<my; y++ )
	for( int x=0; x<mx; x++ )
	{
		minV = INT_MAX;
		for( int dy=-r; dy<=r; dy++ )
		for( int dx=-r; dx<=r; dx++ ) {
			if( dx*dx+dy*dy>r2 ) continue;
			currV = a.safeGet(x+dx, y+dy );
			if( currV < minV ) minV=currV;
		}
		this->put(x,y, minV);
	}
}

void matr::maxNeib( matr &a, int r ) // this[x,y] :=   a    r
{
	a.test();
	int r2 = r*r;
	setSize( a.Mx(), a.My() );
	int maxV, currV;
	for( int y=0; y<my; y++ )
	for( int x=0; x<mx; x++ )
	{
		maxV = INT_MIN;
		for( int dy=-r; dy<=r; dy++ )
		for( int dx=-r; dx<=r; dx++ ) {
			if( dx*dx+dy*dy>r2 ) continue;
			currV = a.safeGet(x+dx, y+dy );
			if( currV > maxV ) maxV=currV;
		}
		this->put(x,y, maxV);
	}
}
/////////////////////////////////////////////////////////////////
int	matr::minNeib(int x0, int y0, int &x1, int &y1)// (x1,y1):=(x0,y0)     a
{
	int v, vM = INT_MAX;
	for( int y=y0-1; y<=y0+1; y++  )
	for( int x=x0-1; x<=x0+1; x++  )
	{
		if( x==x0 && y==y0 ) continue;
		v = this->safeGet(x,y);
		if( v<vM ) {vM=v; x1=x; y1=y; }
	}
	return vM;
}
int	matr::maxNeib(int x0, int y0, int &x1, int &y1)// (x1,y1):=(x0,y0)     a
{
	int v, vM = INT_MIN;
	for( int y=y0-1; y<=y0+1; y++ )
	for( int x=x0-1; x<=x0+1; x++ )
	{
		if( x==x0 && y==y0 ) continue;
		v = this->safeGet(x,y);
		if( v>vM ) {vM=v; x1=x; y1=y; }
	}
	return vM;
}
/////////////////////////////////////////////////////////////////
bool matr::localMax(int x, int y, int h)	//   (x,y)      (2h+1)*(2h+1)
{
	if (x <= h || x >= (mx - h - 1) || y <= h || y >= (my - h - 1)) return false;
	int a0 = get(x, y);
	for (int y1 = y - h; y1 <= y + h; y1++)
		for (int x1 = x - h; x1 <= x + h; x1++) {
			if (x == x1 && y == y1) continue;
			if (get(x1, y1) >= a0) return false;
		}
	return true;
}
bool matr::localMin(int x, int y, int h)	//   (x,y)      (2h+1)*(2h+1)
{
	if (x <= h || x >= (mx - h - 1) || y <= h || y >= (my - h - 1)) return false;
	int a0 = get(x, y);
	for (int y1 = y - h; y1 <= y + h; y1++)
		for (int x1 = x - h; x1 <= x + h; x1++) {
			if (x == x1 && y == y1) continue;
			if (get(x1, y1) <= a0) return false;
		}
	return true;
}
/////////////////////////////////////////////////////////////////
bool matr::emptyBorder(int x, int y, int r)	//      2*r+1 ?
{
	//int d = 2 * r;
	for (int h = -r; h < r; h++) {
		if (safeGet(x + h, y - r) != 0) return false;
		if (safeGet(x + h, y + r) != 0) return false;
		if (safeGet(x - r, y + h) != 0) return false;
		if (safeGet(x + r, y + h) != 0) return false;
	}
	return true;
}
/////////////////////////////////////////////////////////////////
void matr::clearSmall(int r)			//   
{
	int mx1 = mx - r, my1 = my - r;
	int r2 = 2 * r + 1;
	for (int y = r; y < my1; y++)
	for (int x = r; x < mx1; x++) {
		if (get(x, y) == 0) continue;
		if (emptyBorder(x, y, r)) {
			if (r == 1) put(x, y, 0);
			else fillBlock(0, x - r, y - r, r2, r2);
		}
	}

}
void matr::clearSmall()				//   
{
	clearSmall(1);
	clearSmall(2);
	clearSmall(2);
	clearSmall(3);
	clearSmall(4);
	clearSmall(2);
}
/////////////////////////////////////////////////////////////////
ldouble  matr::MSE( matr &a )  // Mean Square Error ( ) (this,a)
{
	this->testSameSize(a);
	ldouble nois=0;
	int diff;
	for( int i=0; i<getSize(); i++ )
	{
		diff = v[i]-a.v[i];
		nois += diff*diff;
	}
	return sqrt( nois/(mx*my) );
}
/////////////////////////////////////////////////////////////////
ldouble  matr::SSIM(matr &a)					// structural similarity (SSIM) index (this,a)
{
	const ldouble c1 = 2.55*2.55;				//   SSIM
	const ldouble c2 = 9 * c1;
	this->testSameSize(a);
	int n = getSize();
	ldouble st = 0, sa=0;						//  this  a
	for (int i = 0; i<n; i++)
	{
		st += v[i];
		sa += a.v[i];
	}
	ldouble midT = st / n, midA = sa / n;
	st = sa = 0;
	ldouble ss = 0;
	for (int i = 0; i<n; i++)
	{
		double dt = v[i] - midT;				//  this
		double da = a.v[i] - midA;				//  a
		st += dt*dt;
		sa += da*da;
		ss += dt*da;
	}
	ldouble Dt = st / n;						//  this
	ldouble Da = sa / n;						//  a
	ldouble Ds = ss / n;						//  (this,a)
	//printf("%10.5f\n", 2 * midT*midA);
	//printf("%10.5f\n", 2 * Ds);
	//printf("%10.5f\n", midT*midT + midA*midA);
	//printf("%10.5f\n", Dt + Da);

	return (2*midT*midA+c1)*(2*Ds+c2)/((midT*midT + midA*midA+c1)*(Dt + Da+c2));
}
ldouble  matr::DSSIM(matr &a)					// =(1-SSIM)/2									[ 1..0]
{
	return 0.5 - 0.5*SSIM(a);
}
/////////////////////////////////////////////////////////////////
static ldouble Fourier_vi(int i,int x,int n)
{
    if (i==0)
    return sqrt(1./n);
    return cos((((2*x+1)*i*dPi))/(2*n))*(sqrt(2.0/n));

}

static ldouble Fourier_vivj(int i,int j,int x,int y, int n,int m)
{
    return Fourier_vi(i,x,n)*Fourier_vi(j,y,m);
}
ldouble	matr::Fourier(int i, int j)				// ij-Fourier coefficient of hole matrix
{
    ldouble f=0;
    for(int y=0;y<my;y++)
    for(int x=0;x<mx;x++)
		f+=get(x,y)*Fourier_vivj(i,j,x,y,mx,my);
    return f;
}
ldouble	matr::Fourier(int i, int j, int x0, int y0, int dx, int dy)// ij-Fourier coefficient of block of matrix
{
    ldouble f=0;
    for(int y=0;y<dy;y++)
    for(int x=0;x<dx;x++)
		f+=get(x0+x,y0+y)*Fourier_vivj(i,j,x,y,dx,dy);
    return f;
}


/////////////////////////////////////////////////////////////////
// RGB
/////////////////////////////////////////////////////////////////
//  
//      r,g,b,a
int	matr::getR( int x, int y ){	return  get(x,y)      & 255; }
int	matr::getG( int x, int y ){	return (get(x,y)>> 8) & 255; }
int	matr::getB( int x, int y ){	return (get(x,y)>>16) & 255; }
int	matr::getA( int x, int y ){	return (get(x,y)>>24) & 255; }

/////////////////////////////////////////////////////////////////
void matr::toRGB( matr &r, matr &g, matr &b )	// this    (r,g,b)
{
	r.setSize( mx, my );
	g.setSize( mx, my );
	b.setSize( mx, my );
	unsigned char *p  = (unsigned char *)v;
	int *pr = r.v;
	int *pg = g.v;
	int *pb = b.v;
	for( int i=mx*my; i>0; i-- )
	{
		*pr ++ = *p ++;
		*pg ++ = *p ++;
		*pb ++ = *p ++;
		p ++;
	}
}

/////////////////////////////////////////////////////////////////
void matr::fromRGB( matr &r, matr &g, matr &b )	// this    (r,g,b)
{
	r.testSameSize(g);
	r.testSameSize(b);
	this->setSizeClear( r.Mx(), r.My() );

	for( int y =0; y <my; y++ )
	for( int x =0; x <mx; x ++ ) {
		int ir = r.get(x,y);
		int ig = g.get(x,y);
		int ib = b.get(x,y);

		ir = ir<0? 0 : ir>255? 255 : ir;
		ig = ig<0? 0 : ig>255? 255 : ig;
		ib = ib<0? 0 : ib>255? 255 : ib;

		put( x, y, ir + (ig<<8) + (ib<<16));
	}
}

/////////////////////////////////////////////////////////////////
void matr::toYUV  ( matr &y, matr &u, matr &v )	// YUV444, ITU-601
{
	y.setSize( mx, my );
	u.setSize( mx, my );
	v.setSize( mx, my );
	int ir, ig, ib;
	int iy, iu, iv;
	for( int yy=0; yy<my; yy++ )
	for( int x =0; x <mx; x ++ ) {
		ir = getR(x,yy);
		ig = getG(x,yy);
		ib = getB(x,yy);

		iy = ( 19595*ir +38470*ig + 7471*ib         )>>16;
		iu = (-11058*ir -21710*ig +32768*ib +8388608)>>16;
		iv = ( 32768*ir -27439*ig - 5329*ib +8388608)>>16;

		y.put( x,yy, iy);
		u.put( x,yy, iu);
		v.put( x,yy, iv);
	}

}

/////////////////////////////////////////////////////////////////
void matr::fromYUV( matr &y, matr &u, matr &v )	// YUV444, ITU-601
{
	y.testSameSize(u);
	y.testSameSize(u);
	this->setSizeClear( y.Mx(), y.My() );

	int ir, ig, ib;
	int iy, iu, iv;
	for( int yy=0; yy<my; yy++ )
	for( int x =0; x <mx; x ++ ) {
		iy = y.get(x,yy);
		iu = u.get(x,yy) - 128;
		iv = v.get(x,yy) - 128;

		ir = (65536*iy          + 91881*iv)>>16;
		ig = (65536*iy- 22553*iu- 46802*iv)>>16;
		ib = (65536*iy+116130*iu          )>>16;

		put( x,yy, ir + (ig<<8) + (ib<<16));
	}

}
/////////////////////////////////////////////////////////////////
// bmp
/////////////////////////////////////////////////////////////////
typedef unsigned __int16 WORD;
#pragma pack(push, 1)
typedef struct {
	  WORD   bfType;			// 0x4d42 | 0x4349 | 0x5450
	  int    bfSize;			//  
	  int    bfReserved;		// 0
	  int    bfOffBits;			//    ,  54 = 16 + biSize
	  int    biSize;			// the number of bytes required by the structure.  40(BITMAPINFOHEADER) || 108(BITMAPV4HEADER) || 124(BITMAPV5HEADER)
	  int    biWidth;			// the width of the bitmap, in pixels
	  int    biHeight;			// the height of the bitmap, in pixels
	  WORD   biPlanes;			//    1
	  WORD   biBitCount;		// 0 | 1 | 4 | 8 | 16 | 24 | 32
	  int    biCompression;		// BI_RGB | BI_RLE8 | BI_RLE4 | BI_BITFIELDS | BI_JPEG | BI_PNG
	  int    biSizeImage;		// Specifies the size, in bytes, of the image. This may be set to zero for BI_RGB bitmaps
	  int    biXPelsPerMeter;	// the horizontal resolution, in pixels-per-meter, of the target device for the bitmap
	  int    biYPelsPerMeter;   // the vertical resolution, in pixels-per-meter, of the target device for the bitmap
	  int    biClrUsed;			//    (   )
	  int    biClrImportant;	//
} BMPheader;
#pragma pack(pop)

int matr::loadBMP( const char *fname )	//       RGB-
{
	FILE *f = fopen( fname, "rb" );
	if( !f ) { setSize(0,0); return -1; }
	BMPheader bh;	// File header sizeof(BMPheader) = 56
	size_t res;

	//  
	res = fread( &bh, 1, sizeof(BMPheader), f );
	if( res != sizeof(BMPheader) ) { fclose(f); return -1; }

	//  
	if( bh.bfType!=0x4d42 && bh.bfType!=0x4349 && bh.bfType!=0x5450 ) { fclose(f); return -1; }

	//   
	fseek( f, 0, SEEK_END);
	int filesize = ftell(f);
	//    :
	fseek( f, sizeof(BMPheader), SEEK_SET);
	//  
	if( bh.bfSize != filesize ||
	    bh.bfReserved != 0    ||
	    bh.biPlanes   != 1    ||
	   (bh.biSize!=40 && bh.biSize!=108 && bh.biSize!=124)||
	    bh.bfOffBits != 14+bh.biSize ||

	    bh.biWidth <1 || bh.biWidth >20000 ||
	    bh.biHeight<1 || bh.biHeight>20000 ||
	    bh.biBitCount    != 24 ||                         	//     
	    bh.biCompression !=  0								//     
		)
	{
			fclose(f);
			setSize(0,0);
			return -1;
	}
	//    ,  -  (BGR-24),  (mx,my) 
	int mx_ = bh.biWidth;
	int my_ = bh.biHeight;
    int mx3 = (3*mx_+3) & (-4);	// Compute row width in file, including padding to 4-byte boundary
	unsigned char *tmp_buf = new unsigned  char[mx3*my_];	//  
	res = fread( tmp_buf, 1, mx3*my_, f);
	if( (int)res != mx3*my_ ) { delete []tmp_buf; fclose(f); setSize(0,0); return -1; }
	//  
	fclose(f);
	//  mx, my   
	setSize( mx_, my_);

	// Transfer data.  Note source values are in BGR order
	unsigned char *ptr = (unsigned char *) this->v;
	for(int y = my-1; y >= 0; y--) {
		unsigned char *pRow = tmp_buf + mx3*y;
		for(int x=0; x< mx; x++) {
			*ptr++ = *(pRow + 2);
			*ptr++ = *(pRow + 1);
			*ptr++ = *pRow;
			pRow+=3;
			ptr ++;
		}
	}

	delete []tmp_buf;
	return 0;	// OK

}
/////////////////////////////////////////////////////////////////
int matr::loadBMPGrey( const char *fname )	// returns error code
{
	int res = this->loadBMP(fname);
	if( res ) return res;
	// RGB - 
	int ir, ig, ib;
	int iy;
	for( int yy=0; yy<my; yy++ )
	for( int x =0; x <mx; x ++ ) {
		ir = getR(x,yy);
		ig = getG(x,yy);
		ib = getB(x,yy);

		iy = ( 19595*ir +38470*ig + 7471*ib         )>>16;

		put( x,yy, iy);
	}
	return 0;
}
/////////////////////////////////////////////////////////////////
int matr::saveBMP( const char *fname )	//       RGB-
{
	BMPheader bh;	// File header sizeof(BMPheader) = 56
	memset( &bh, 0, sizeof(bh) );
	bh.bfType =0x4d42;	// 'BM'
    // Compute row width in file, including padding to 4-byte boundary
    int mx3 = (3*mx+3) & (-4);
	int filesize = 54 + my*mx3;
	bh.bfSize = filesize;
	bh.bfReserved =  0;
	bh.biPlanes   =  1;
	bh.biSize     = 40;
	bh.bfOffBits  = 14 + bh.biSize;
	bh.biWidth    = mx;
	bh.biHeight   = my;
	bh.biBitCount = 24;
	bh.biCompression= 0;

	FILE *f = fopen( fname, "wb" );
	if( !f ) return -1;
	size_t res;

	//  
	res = fwrite( &bh, 1, sizeof(BMPheader), f );
	if( res != sizeof(BMPheader) ) { fclose(f); return -1; }

	//   
	unsigned char *tmp_buf = new unsigned char[mx3*my];
	// Transfer data.  Note source values are in BGR order
	unsigned char *ptr = (unsigned char *)(this->v);
	for(int y = my-1; y >= 0; y--) {
		unsigned char *pRow = tmp_buf + mx3*y;
		for(int x=0; x< mx; x++) {
			*(pRow + 2) = *ptr++;
			*(pRow + 1) = *ptr++;
			*pRow       = *ptr++;
			pRow+=3;
			ptr++;
		}
	}
	//   
	fwrite( tmp_buf, 1, mx3*my, f );
	fclose(f);
	delete []tmp_buf;

	return 0;	// OK

}
/////////////////////////////////////////////////////////////////
//  k=0,      [0..255]
int matr::saveBMPGrey( const char *fname, int k ) //      k
{
	matr  grey;
	grey.copyFrom(*this);
	if( k > 1) grey.mult(k);
	else if(k==0) {									//  k=0,      [0..255]
		grey.addV( -grey.minimal() );				//   
		grey.mult( INT_MAX/2/grey.maximal());		//    [0..MAX_INT/2]
		grey.div ( (grey.maximal()+127)/255 );		//    [0..k]
	}
	grey.clip(0, 255);
	grey.mult(0x10101);
	return grey.saveBMP(fname);
}
/////////////////////////////////////////////////////////////////
void    matr::loadPart(const char*fname, int x0, int y0, int dx, int dy)	//   
{
	if (x0<0 || y0<0 || dx<1 || dy<0) vError("loadPart: illegal input sizes");
	matr t1;
	t1.loadBMP(fname);
	setSize(dx, dy);
	for (int y = 0; y<dy; y++)
		for (int x = 0; x<dx; x++)
			put(x, y, t1.safeGet(x0 + x, y0 + y));
}
/////////////////////////////////////////////////////////////////
void  matr::loadPartGrey(const char*fname, int x0, int y0, int dx, int dy)//   
{
	if (x0<0 || y0<0 || dx<1 || dy<0) vError("loadPartGrey: illegal input sizes");
	matr t1;
	t1.loadBMPGrey(fname);
	setSize(dx, dy);
	for (int y = 0; y<dy; y++)
		for (int x = 0; x<dx; x++)
			put(x, y, t1.safeGet(x0 + x, y0 + y));
}
/////////////////////////////////////////////////////////////////
int	matr::loadTifSimple(const char *fname)	//   tiff-   -, returns error code
{
	setSize(0, 0);
	FILE *tif = fopen(fname, "rb");
	if (!tif) return 1;

	char tif_header[8];
	if (fread(tif_header, 1, 8, tif) != 8) return 1;
	if (tif_header[0] != 0x49 ||
		tif_header[1] != 0x49 ||
		tif_header[2] != 0x2a ||
		tif_header[3] != 0x00) return 2;
	if (tif_header[4] != 0x08 ||
		tif_header[5] != 0x00 ||
		tif_header[6] != 0x00 ||
		tif_header[7] != 0x00) return 2;
	unsigned short numberOfDir;
	if (fread(&numberOfDir, 2, 1, tif) != 1) return 3;
	struct {
		unsigned short tag;
		unsigned short type;
		unsigned count;
		unsigned off;
	} DE;
	int mx = -1, my = -1, bitPixel = -1;
	for (unsigned i = 0; i < numberOfDir; i++) {
		if (fread(&DE, 12, 1, tif) != 1) return 4;
		if (DE.tag == 256) mx = DE.off;
		if (DE.tag == 257) my = DE.off;
		if (DE.tag == 258) {
			bitPixel = DE.off;
			if (bitPixel != 8 && bitPixel != 16) return 8; // vError("tif_er 08; bitPixel = %d", bitPixel);
		}
		if (DE.tag == 259) if (DE.off != 1) return 5; // vError("tif_er 05; File is compressed");
		if (DE.tag == 278) if (DE.off != my) return 6; // vError("tif_er 06; File is too complicated");
		if (DE.tag == 320) return 7; // vError("tif_er 07; Colors!");
	}

	//tag =    256, tape =      4, count =         1, off =     16000
	//tag =    257, tape =      4, count =         1, off =       576
	//tag =    258, tape =      3, count =         1, off =        16
	//tag =    259, tape =      3, count =         1, off =         1
	//tag =    262, tape =      3, count =         1, off =         1
	//tag =    273, tape =      4, count =         1, off =       174
	//tag =    278, tape =      4, count =         1, off =       576
	//tag =    279, tape =      4, count =         1, off =  18432000
	//tag =    282, tape =      5, count =         1, off =       158
	//tag =    283, tape =      5, count =         1, off =       166
	//tag =    284, tape =      3, count =         1, off =         1
	//tag =    296, tape =      3, count =         1, off =         3
	//-------------------------------------------------------------

	//tag =    256, tape =      4, count =         1, off =       500
	//tag =    257, tape =      4, count =         1, off =       500
	//tag =    258, tape =      3, count =         1, off =         8
	//tag =    259, tape =      3, count =         1, off =         1
	//tag =    262, tape =      3, count =         1, off =         1
	//tag =    273, tape =      4, count =         1, off =       122
	//tag =    278, tape =      4, count =         1, off =       500
	//tag =    279, tape =      4, count =         1, off =    250000
	//tag =    284, tape =      3, count =         1, off =         1


	unsigned size = mx*my;
	this->setSize(mx, my);
	if (bitPixel == 8) {
		fread(&DE, 1, 4, tif);
		unsigned char *buf = new unsigned char[size];
		if (fread(buf, 1, size, tif) != size) { delete[]buf; return 8; }
		unsigned char *p = buf;
		for (int y = 0; y < my; y++)
			for (int x = 0; x < mx; x++)
				this->put(x, y, *p++);
		delete[]buf;
	}
	else {
		fread(&DE, 1, 12, tif);
		fread(&DE, 1, 8, tif);
		unsigned short *buf = new unsigned short[size];
		if (fread(buf, 2, size, tif) != size) { delete[]buf; return 8; }
		unsigned short *p = buf;
		for (int y = 0; y < my; y++)
			for (int x = 0; x < mx; x++)
				this->put(x, y, *p++);
		delete[]buf;
	}
	fclose(tif);
	return 0;
}
/////////////////////////////////////////////////////////////////
ldouble matr::MSE_rgb( matr &a )					// Mean Square Error (this,a)  RGB- 
{
	this->testSameSize(a);
	ldouble nois=0;
	int ti, ai, dr, dg, db;
	for( int i=0; i<getSize(); i++ )
	{
		ti = v[i];
		ai = a.v[i];

		dr = ( ti      & 255) - (ai       & 255);
		dg = ((ti>> 8) & 255) - ((ai>> 8) & 255);
		db = ((ti>>16) & 255) - ((ai>>16) & 255);
		nois += dr*dr + dg*dg + db*db;
	}
	return sqrt( nois/(3*mx*my) );
}
/////////////////////////////////////////////////////////////////
void	matr::roundSize( int k)	//      k
{
	int mxk = k*(mx/k);
	int myk = k*(my/k);
	if( mxk==mx && myk==my ) return;	//     ,  
	int *v0 = v;		//   
	v = new int[mxk*myk];
	for( int y=0; y<myk; y++ )
	for( int x=0; x<mxk; x++ )
		v[x+y*mxk] = v0[x+y*mx];

	delete []v0;
	mx = mxk;
	my = myk;

}
/////////////////////////////////////////////////////////////////
void	matr::transpose( matr &a)				// this := transpose(a)
{
	setSize(a.My(), a.Mx() );
	for( int y=0; y<my; y++ )
	for( int x=0; x<mx; x++ )
		put(x,y, a.get(y,x));
}
/////////////////////////////////////////////////////////////////
//  hist    
void matr::historgrammK( int k0, int k1, int *hist ) // hist:=   k0..k1 
{
	int x,y, v;
	for( y=k0; y<=k1; y++ ) hist[y-k0]=0;

	for( y=0; y<my; y++ )
	for( x=0; x<mx; x++ )
	{
		v=get(x,y);
		if( v>=k0 && v<=k1 ) hist[v-k0]++;
	}
}// void matr::historgrammK( int k0, int k1, int *hist ) // hist:=   k0..k1 
/////////////////////////////////////////////////////////////////
// v0:  ,   <v0  p0.    v0
// v1:  ,   >v1  p1.    v1
// , :
// matr.Quantile( 0.05, v0, 0.03, v1);
//   (v0,v1) , :
//   ,  v0  5%
//   ,  v1  3%
void matr::Quantile( ldouble p0, int &v0, ldouble p1, int &v1 )	// (v0,v1):= (p0,p1)     a
{
	int x,y, v, cnt, vMin=INT_MAX, vMax=INT_MIN;
	for( y=0; y<my; y++ )
	for( x=0; x<mx; x++ )
	{
		v=get(x,y);
		if( v<vMin ) vMin = v;
		if( v>vMax ) vMax = v;
	}
	//  vMin,vMax -      
	if( vMax-vMin > 1000000 || vMax==vMin ) vError("matr::Quantile,vMin=%d, vMax=%d", vMin, vMax);
	int *hist = new int[vMax-vMin+1];	//   
	this->historgrammK(vMin, vMax, hist);
	//  

	int ip0 = round(p0*mx*my);
	int ip1 = round(p1*mx*my);
	//    v0 ..    v0   ip0.
	cnt = 0;
	v0  = vMax;
	for( v=vMin; v<=vMax; v++ )
	{
		cnt += hist[v-vMin];
		if(cnt>=ip0) {v0=v; break;}
	}

	//    v1 ..    v1   ip1.
	cnt = 0;
	v1  = vMin;
	for( v=vMax; v>=vMin; v-- )
	{
		cnt += hist[v-vMin];
		if(cnt>=ip1) {v1=v; break;}
	}
	delete []hist;
}
//-----------------------------------------------------------------------------
ldouble matr::entropy() // : sum p[i]*log(p[i])
{
	int mi = minimal();
	int ma = maximal();
	int *hist = new int[ma-mi+1];
	historgrammK(mi, ma, hist);
	//   hist     a,
	//   hist[i] =     mi+i   a
	ldouble sz1 = 1./(mx*my);
	ldouble entr = 0;
	for( int i=0; i<ma-mi+1; i++ )
	{
		ldouble pi = hist[i]*sz1;
		if( pi>0)
			entr -= pi*log(pi);
	}
	delete []hist;
	return entr/log(2.);
}
//-----------------------------------------------------------------------------
void matr::setRandom(int mx, int my, int maxV)//    0  maxV-1
{
	setSize(mx, my);
	for (int i = 0; i<getSize(); i++) v[i] = hRandom(maxV);
}
//-----------------------------------------------------------------------------
/////////////////////////////////////////////////////////////////
//===========================================================================================
//   
//===========================================================================================

//-----------------------------------------------------------------------------
// a[x,y] := a*x + b*y + c    minV  maxV
//
void matr::setLinear( int mx, int my, int minV, int maxV, ldouble ax, ldouble ay, ldouble a0 ) // a[x,y]:=a*x+b*y+c
{
	setSize(mx,my);
	int x,y;
	for( y=0; y<my; y++ )
	for( x=0; x<mx; x++ )
	{
		ldouble d   = round( ax*x + ay*y + a0 );
		if( d<minV ) d=minV;
		if( d>maxV ) d=maxV;
		put(x,y, round(d) );
	}
}
//===========================================================================================
//   
//===========================================================================================

void mix( matr &res, matr &a, matr &b, ldouble al ) // res:= al*a + (1-al)*b
{
	int mx=a.Mx(), my=a.My();
	if( b.Mx() != mx || b.My() != my ) { res.setSize(0,0); return; }
	res.setSize(mx,my);
	ldouble be = 1-al;
	for( int y=0; y<my; y++ )
	for( int x=0; x<mx; x++ )
		res.put(x,y, round(al*a.get(x,y) + be*b.get(x,y) ) );
}
//-----------------------------------------------------------------------------


