// iMatrix.h: matrix with integer coefficients
// 2006-11-06
//
#ifdef IMATRIX
#error Secondary include
#endif
#define IMATRIX

//-----------------------------------------------------------------------------
#define EDGE_ITERATOR
class edgeIterator {	//    
private:
	int x0, y0, x1, y1;
	int deltaX, deltaY, signX, signY, error;
public:
	~edgeIterator() {};				//   
	edgeIterator(int x0_, int y0_, int x1_, int y1_);
	edgeIterator(int *P0, int *P1);
	bool next();					// ,   
	int X() { return x0; };
	int Y() { return y0; };
};
// :
//	matrix a(800,600)
//	edgeIterator ei(20, 120, 120, 30);
//	do a(ei.Y(), ei.X()) = 255;
//	while (!ei.next());


class matr{
protected:
	int mx, my;		// dimension
	int *v;			// matrix

public:
	// constructors/destructor
	matr(int mx_=0, int my_=0);
	matr( matr &a );		// vError("matr::matr(matr &a): copy constructor illegal");
	~matr();
	// memory, sizes
	void	release();		// release memory
	void	clear();		//   
	//    ,    ,  ,   
	void	setSize		(int mx_, int my_);	//  
	void	setSizeClear(int mx_, int my_);	//     
	void	copyFrom(matr &a);
	// tests
	void	test();
    void	testSameSize(matr &a); 
	void	testIndexes(int x, int y);		// 0<=x,y<max
	bool    isIndexes(int x, int y);		// ?0<=x,y<max
	void	normaIndexes(int &x, int &y);	//     0<=x,y<max
	// get/put
	int		Mx() { return mx; }
	int		My() { return my; }
	int		getSize(){ return mx*my; }
	int     *getElement(int x, int y);
	int		get( int x, int y );
	void	put( int x, int y, int value );
	void	inc( int x, int y, int value );		//     value
	int     &operator () (int, int);

	int		safeGet( int x, int y );			//     ,   
	void	safePut( int x, int y, int value ); //  ,       
	void	set( int value );					//     
//protected:
	void	fillLine  ( int y0, int v );		//    v
	void	fillColumn( int x0, int v );		//    v
	void	copyLine  ( int yDst, int ySrc );	//  
	void	copyColumn( int xDst, int xSrc );	//  
//public:
	// interpolation
	ldouble	getInterp ( ldouble x, ldouble y, int k=1 ); // universal interpolation k=0|1|3
	ldouble	getInterp1( ldouble x, ldouble y );	// bilinear get( x, y )
	ldouble	getInterp3( ldouble x, ldouble y );	// bicubic get( x, y )

	// in/out
	void	saveText( const char *fname, const char *msg=NULL, const char *format = " %5d" );
	void	appendText(const char *fname, const char *msg = NULL, const char *format = " %5d");
	void	saveTextI(const char *fname, const char *msg = NULL, const char *format = " (%4d,%4d=>%6d)"); //   
	void	saveTextB(const char *fname, int x0, int y0, int dx, int dy, const char *msg = NULL, const char *format = " (%4d,%4d=>%6d)"); //   
	void	loadText( const char *fname );
	// fill, copy
	void	copyBlock( matr &a, int x0, int y0, int dx, int dy );
	void	copyBlock( matr &a, int xSrc, int ySrc, int xDst, int yDst, int dx, int dy );// this:=a[]
	void	fillBlock( int v, int x0, int y0, int dx, int dy );
	void	fillCircle( int v,int xr, int yr, int r );
	void	fillOuter( int x0, int y0, int x1, int y1, int v); //    (x0,y0, x1,y1)  v
	void	swap( matr &a );					// this <-> a
	// min, max, count, change
	int		minimal(int &x_, int &y_);			// minimal value
	int		minimal();							// minimal value
	int		maximal(int &x_, int &y_);			// maximal value
	int		maximal();							// maximal value

	int		minLin(int &x_, int y_);			// minimal value in line y, (x,y) - point
	int		maxLin(int &x_, int y_);			// maximal value in line y, (x,y) - point
	int		minCol(int x_, int &y_);			// minimal value in column x, (x,y) - point
	int		maxCol(int x_, int &y_);			// maximal value in column x, (x,y) - point


	ldouble mid();								// mid value
	void	drawEdge(int x0, int y0, int x1, int y1, int c); //   [(x0,y0),(x1,y1)]  c
	ldouble midEdge(int x0, int y0, int x1, int y1); // mid value on the edge (x0,y0)-(x1,y1)
	ldouble midEdg2(int x0, int y0, int x1, int y1); // mid*sqrt(len)
	int		minEdge(int x0, int y0, int x1, int y1, int &x2, int &y2); // min value on the edge (x0,y0)-(x1,y1)
	int		maxEdge(int x0, int y0, int x1, int y1, int &x2, int &y2); // max value on the edge (x0,y0)-(x1,y1)
	ldouble midBox(int x0, int y0, int x1, int y1); // mid value on the box (x0,y0)-(x1,y1)

	int		countVal( int val );				//  ,  v
	int		countValI( int val0, int val1 );	//   x: val0<=x<val1
	int		changeVal(int v1, int v2);			//  v1  v2.   
	int		changeVal2(int v1, int v2, int v3);	//     [v1..v2]  v3.   
	void	clip( int min_, int max_ );			// ,  min   min, max - 
	void	normalize( int k );					//       [0..k]
	void	Mod( int k );						//     mod k (0..k-1)

	// algebra
	void	mult( int k);						// this *= k
	void	mult( matr &a);						// this(x,y) *= a(x,y) 
	void	div ( int k);						// this /= k
	void	add( matr &a);						// this += a
	void	add( matr &a, int k);				// this += a*k
	void	addV( int r);						// this[*,*] += r
	void	sub( matr &a);						// this -= a
	void	addMult( matr &a, int k);			// this += a*k
	void    mask(matr &a);						//  a(x,y)!=0,  this(x,y):=a(x,y)
	void    mask(matr &a, int x0, int y0);		//  a(x,y)!=0,  this(x0+x,y0+y):=a(x,y)
	void	addB(matr &a);						// this += a 

	int		sumRow( int y);						//   y- 
	int		sumColumn( int x);					//   x- 
	ldouble CSprite(matr &spr, int x, int y);	//      

	// partial derivative
	int		dx( int x, int y);					// 2*da/dx in the point
	int		dy( int x, int y);					// 2*da/dy in the point
	int		dxx(int x, int y);					// d2a/dxdx in the point
	int		dxy(int x, int y);					// d2a/dxdy in the point
	int		dyy(int x, int y);					// d2a/dydy in the point
	double	Rinvar (int x, int y);				// R-  
	int		Rinvars(int x, int y);				// (-1|0|1)  R-  
	ldouble	l2Grad( int x, int y);				// length of gradient in the point
	int		mGrad(int x, int y, int h = 1);		// morphological gradient in a point (max-min amoung (2*h+1)*(2*h+1) neigbour points)
	int		variat(int x, int y, int h = 1);	//         (2*h+1)*(2*h+1)

	double midVar();							//    

	void	diffX(matr &a);						// this:= 2*da/dx 
	void	diffY(matr &a);						// this:= 2*da/dy 
	void	diffXX(matr &a);					// this:= d^2a/dxdx  
	void	diffXY(matr &a);					// this:= d^2a/dxdy  
	void	diffYY(matr &a);					// this:= d^2a/dydy  
	void 	l2Grad(matr &a);					// this:= length of gradient in each point(a)
	void	mGrad(matr &a, int h = 1);			// this:= morphological gradient in each point(a) (max-min amoung (2*h+1)*(2*h+1) neigbour points)
	void	variat(matr &a, int h = 1);			// this:=         (2*h+1)*(2*h+1)

	void	smooth( int r, int kind=1 );		// smoothing with radius r
												// kind=0   .  
												// kind=1   
												// kind=2   

	void	decrease( matr &a, int k );			// this := decrease size k times(a)
	void	resize( matr &a, int mx1, int my1);	// this := resize(a)  (mx,my)
	void	minNeib( matr &a, int r );			// this[x,y] := minimal value in r-vicinity 
	void	maxNeib( matr &a, int r );			// this[x,y] := maximal value in r-vicinity

	int		minNeib(int x0, int y0, int &x1, int &y1);// (x1,y1):=Neibour(x0,y0) with minimal a
	int		maxNeib(int x0, int y0, int &x1, int &y1);// (x1,y1):=Neibour(x0,y0) with maximal a

	bool    localMax(int x, int y, int h = 1);	//   (x,y)      (2h+1)*(2h+1)
	bool    localMin(int x, int y, int h = 1);	//   (x,y)      (2h+1)*(2h+1)
	bool    emptyBorder(int x, int y, int r);	//      2*r+1 ?
	void    clearSmall(int r);					//   
	void    clearSmall();						//   

	// 
	ldouble  MSE(matr &a);						// Mean Square Error (this,a)
	ldouble  SSIM(matr &a);						// structural similarity (SSIM) index (this,a)  [-1..1]
	ldouble  DSSIM(matr &a);					// =(1-SSIM)/2									[ 1..0]


	// coefficients of Fourier transformations
	ldouble	Fourier(int i, int j);				// ij-Fourier coefficient of a whole matrix 
	ldouble	Fourier(int i, int j, int x0, int y0, int dx, int dy);// ij-Fourier coefficient of block of matrix 

	// RGB, YUV
	//      r,g,b,a
	int		getR( int x, int y );
	int		getG( int x, int y );
	int		getB( int x, int y );
	int		getA( int x, int y );

	void	toRGB  ( matr &r, matr &g, matr &b );
	void	fromRGB( matr &r, matr &g, matr &b );

	void	toYUV  ( matr &y, matr &u, matr &v );	// YUV444, ITU-601
	void	fromYUV( matr &y, matr &u, matr &v );	// YUV444, ITU-601

	// bmp
	int		loadBMP    ( const char *fname );	//       RGB-
	int		loadBMPGrey( const char *fname );	// returns error code
	int		saveBMP    ( const char *fname );	//       RGB-
	int		saveBMPGrey( const char *fname, int k=1 ); //          k
	void    loadPart   ( const char*fname, int x0, int y0, int dx, int dy);	//   
	void    loadPartGrey(const char*fname, int x0, int y0, int dx, int dy);//   

	int		loadTifSimple(const char *fname);	//   tiff-   -, returns error code

	ldouble  MSE_rgb( matr &a );				// Mean Square Error (this,a)  RGB- 
	//
	void	roundSize( int k);					//      k
	void	transpose( matr &a);				// this := transpose(a)
	void	historgrammK( int k0, int k1, int *hist ); // hist:=   k0..k1 
	void	Quantile ( ldouble p0, int &v0, ldouble p1, int &v1 );	// (v0,v1):= (p0,p1)     a
	ldouble	entropy  ();						// : sum p[i]*log(p[i])
	void    setLinear(int mx, int my, int minV, int maxV, ldouble ax, ldouble ay, ldouble a0);// a[x,y]:=a*x+b*y+c
	void    setRandom(int mx, int my, int maxV);//    0  maxV-1

};

//===========================================================================================
//   
//===========================================================================================

void mix( matr &res, matr &a, matr &b, ldouble al ); // res:= al*a + (1-al)*b


//===========================================================================================
//   
//===========================================================================================


