// iMatrix3.cpp:     matr
// 2006-11-07
//
#ifdef _MSC_VER
  #pragma warning(disable : 4996)
#endif

#ifdef __GNUC__
#pragma GCC diagnostic ignored "-Wwrite-strings"
#pragma GCC diagnostic ignored "-Wformat"
#pragma GCC diagnostic ignored "-Wsign-compare"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <math.h>

#include "vUtil.h"
#include "iMatrix.h"		//  
#include "iMatrix3.h"		// 3-


////////////////////////////////////////////////////
// constructors/destructors
////////////////////////////////////////////////////
//   mx*my.
matr3::matr3(int mx_, int my_){ setSize(mx_, my_); }
void matr3::release(){ r.release(); g.release(); b.release(); }
matr3::~matr3() { this->release(); }
void matr3::clear() { r.clear(); g.clear(); b.clear(); }
void matr3::setSize( int mx_, int my_){ r.setSize(mx_,my_); g.setSize(mx_,my_); b.setSize(mx_,my_); }
void matr3::setSizeClear( int mx_, int my_){ r.setSizeClear(mx_,my_); g.setSizeClear(mx_,my_); b.setSizeClear(mx_,my_); }
void matr3::copyFrom(matr3 &a){ r.copyFrom(a.r); g.copyFrom(a.g); b.copyFrom(a.b); }
void matr3::swap(matr3 &a) { r.swap(a.r);  g.swap(a.g); b.swap(a.b); }
void matr3::test() { r.testSameSize(g); r.testSameSize(b); }
void matr3::testSameSize( matr  &a  ){ test(); r.testSameSize(a); }
void matr3::testSameSize( matr3 &a  ){ test(); a.test(); r.testSameSize(a.r); }
void matr3::testIndexes(int x, int y){ r.testIndexes(x,y); }
void matr3::normaIndexes(int &x, int &y){ r.normaIndexes(x,y); }
///////////////////////////////////////////////////////////
// get/put
void matr3::put( int x, int y, int vr, int vg, int vb ){ r.put(x,y,vr); g.put(x,y,vg); b.put(x,y,vb); }
int  matr3::get( int x, int y ) { return rgb_int( r.get(x,y), g.get(x,y), b.get(x,y)); }	// r + 256*g + 256^2*b
void matr3::inc( int x, int y, int vr, int vg, int vb ){ r.inc(x,y,vr); g.inc(x,y,vg); b.inc(x,y,vb); }
void matr3::safePut( int x, int y, int vr, int vg, int vb ){ r.safePut(x,y,vr); g.safePut(x,y,vg); b.safePut(x,y,vb); }
void matr3::set( int vr, int vg, int vb ){ r.set(vr); g.set(vg); b.set(vb); }

////////////////////////////////////////////////////////////////////////////
// algebra
void matr3::add ( matr3 &a ){ r.add(a.r); g.add(a.g); b.add(a.b); };	// this += a
void matr3::add ( matr3 &a, int k){ r.add(a.r,k); g.add(a.g,k); b.add(a.b,k); };	// this += a*k
void matr3::mult( int k){ r.mult(k); g.mult(k); b.mult(k); };	// this *= k
void matr3::div ( int k){ r.div (k); g.div (k); b.div (k); };	// this /= k
void matr3::addV( int ar, int ag, int ab){ r.addV(ar); g.addV(ag); b.addV(ab); };	// this.r[*,*] += ar, this.g[*,*] += ag, this.b[*,*] += ab

////////////////////////////////////////////////////////////////////////////
int matr3::mGrad( int x, int y) // .  : . .      
{
	int dist2 = 0;
	int r0,g0,b0, dr,dg,db, d2;
	r0 = r.get(x,y);
	g0 = g.get(x,y);
	b0 = b.get(x,y);

	for( int y1=y-1; y1<=y+1; y1++ )
	for( int x1=x-1; x1<=x+1; x1++ )
	{
		if( x1==x && y1==y ) continue;
		dr = r.safeGet(x1,y1)-r0;
		dg = g.safeGet(x1,y1)-g0;
		db = b.safeGet(x1,y1)-b0;

		//d2 = dr*dr + dg*dg + db*db;
		d2 = iabs(dr) + iabs(dg) + iabs(db);
		if( d2>dist2 ) dist2=d2;
	}
	return dist2;
}
////////////////////////////////////////////////////////////////////////////
ldouble matr3::dist(int x1, int y1, int x2, int y2)//    
{
	ldouble dr = r.get(x1,y1)-r.get(x2,y2);
	ldouble dg = g.get(x1,y1)-g.get(x2,y2);
	ldouble db = b.get(x1,y1)-b.get(x2,y2);
	return sqrt((ldouble)(dr*dr + dg*dg + db*db));
}
////////////////////////////////////////////////////////////////////////////
void matr3::copyBlock( matr3 &a, int xSrc, int ySrc, int xDst, int yDst, int dx, int dy )// this:=a[]
{
	this->r.copyBlock(a.r, xSrc, ySrc, xDst, yDst, dx, dy);
	this->g.copyBlock(a.g, xSrc, ySrc, xDst, yDst, dx, dy);
	this->b.copyBlock(a.b, xSrc, ySrc, xDst, yDst, dx, dy);
}
////////////////////////////////////////////////////////////////////////////
// in/out
void matr3::saveText( const char *fname, const char *msg, char *format )
{
	this->test();
	FILE *f = fopen( fname, "w" );
	if(!f) vError( "saveText: file not open" );
	fprintf( f, "%4d %4d %s\n", Mx(), My(), msg );
	for( int y=0; y<My(); y++ )
	{
		for( int x=0; x<Mx(); x++ )
		{
			//fprintf( f, format, this->get(x,y) );
			fprintf( f, format, x, y, r.get(x,y), g.get(x,y), b.get(x,y) );
		}
		fprintf( f, "\n");
	}
	fclose(f);
}

void matr3::smooth( int rad, int kind ){ r.smooth(rad, kind); g.smooth(rad, kind); b.smooth(rad, kind); }
void matr3::decrease( matr3 &a, int k ){ r.decrease(a.r, k);  g.decrease(a.g, k);  b.decrease(a.b, k); }
void matr3::resize(matr3 &a,int mx1,int my1){ r.resize(a.r,mx1,my1);  g.resize(a.g,mx1,my1);  b.resize(a.b,mx1,my1); }
void matr3::diffX   ( matr3 &a        ){ r.diffX(a.r);        g.diffX(a.g);        b.diffX(a.b); }
void matr3::diffY   ( matr3 &a        ){ r.diffY(a.r);        g.diffY(a.g);        b.diffY(a.b); }

ldouble matr3::MSE( matr3 &a ) {
	ldouble dr = r.MSE(a.r);
	ldouble dg = g.MSE(a.g);
	ldouble db = b.MSE(a.b);
	return sqrt((dr*dr + dg*dg + db*db)/3);
}
//-----------------------------------------------------------------------------
ldouble  matr3::SSIM(matr3 &a)					// structural similarity (SSIM) index (this,a)
{
	const ldouble c1 = 2.55*2.55;				//   SSIM
	const ldouble c2 = 9 * c1;
	this->testSameSize(a);
	int n = getSize();
	ldouble st = 0, sa = 0;						//  this  a
	int *ptr = r.getElement(0, 0);
	int *ptg = g.getElement(0, 0);
	int *ptb = b.getElement(0, 0);

	int *par = a.r.getElement(0, 0);
	int *pag = a.g.getElement(0, 0);
	int *pab = a.b.getElement(0, 0);


	for (int i = 0; i<n; i++)
	{
		st += ptr[i] + ptg[i] + ptb[i];
		sa += par[i] + pag[i] + pab[i];
	}
	ldouble midT = st /(3*n), midA = sa / (3 * n);
	st = sa = 0;
	ldouble ss = 0,dt, da;
	for (int i = 0; i<n; i++)
	{
		dt = ptr[i] - midT;				//  this
		da = par[i] - midA;				//  a
		st += dt*dt;
		sa += da*da;
		ss += dt*da;

		dt = ptg[i] - midT;				//  this
		da = pag[i] - midA;				//  a
		st += dt*dt;
		sa += da*da;
		ss += dt*da;

		dt = ptb[i] - midT;				//  this
		da = pab[i] - midA;				//  a
		st += dt*dt;
		sa += da*da;
		ss += dt*da;
	}
	ldouble Dt = st / (3 * n);						//  this
	ldouble Da = sa / (3 * n);						//  a
	ldouble Ds = ss / (3 * n);						//  (this,a)
	return (2 * midT*midA + c1)*(2 * Ds + c2) / ((midT*midT + midA*midA + c1)*(Dt + Da + c2));
}
ldouble  matr3::DSSIM(matr3 &a)					// =(1-SSIM)/2									[ 1..0]
{
	return 0.5 - 0.5*SSIM(a);
}

//-----------------------------------------------------------------------------

void matr3::toRGB( matr &rgb ){	rgb.fromRGB( r, g, b); }	//      rgb
void matr3::fromRGB( matr &rgb ){ rgb.toRGB( r, g, b); }	//      rgb

int matr3::loadBMP( const char *fname ) {
	matr rgb;
	int res = rgb.loadBMP(fname);
	if( res ) return res;
	this->fromRGB(rgb);
	return 0;
}

int matr3::saveBMP( const char *fname ) {
	matr rgb;
	this->toRGB(rgb);
	return rgb.saveBMP(fname);
}

/////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////
//   [(x0,y0),(x1,y1)]  
void matr3::drawEdge( int x0, int y0, int x1, int y1 )
{
    r.drawEdge(x0,y0,x1,y1, -1);
    g.drawEdge(x0,y0,x1,y1, -1);
    b.drawEdge(x0,y0,x1,y1, -1);
}

//   [(x0,y0),(x1,y1)]  (cr,cb,cb)
void matr3::drawEdge( int x0, int y0, int x1, int y1, int cr, int cg, int cb )
{
    r.drawEdge(x0,y0,x1,y1, cr);
    g.drawEdge(x0,y0,x1,y1, cg);
    b.drawEdge(x0,y0,x1,y1, cb);
}
//===========================================================================================
void matr3::normalize( int k )					//       [0..k]
{
	//r.addV( -r.minimal() );
	//g.addV( -g.minimal() );
	//b.addV( -b.minimal() );				//       -  
	//int ma = imax3( r.maximal(), g.maximal(), b.maximal());
	//div( (ma+k-1)/k );
	r.normalize(k);
	g.normalize(k);
	b.normalize(k);
}
//===========================================================================================
void matr3::Mod( int k )		//     mod k (0..k-1)
{
	this->r.Mod(k);
	this->g.Mod(k);
	this->b.Mod(k);
}
//===========================================================================================
void matr3::GrLen3( matr  &grl )					// grl :=    (a)
{
	this->test();
	int mx=this->Mx(), my=this->My();

	grl.setSize(mx,my);
	int x,y;
	for( y=0; y<my; y++ )
	for( x=0; x<mx; x++ )
	{
		ldouble isq = dSquare(r.dx(x,y)) + dSquare(r.dy(x,y)) +
				     dSquare(g.dx(x,y)) + dSquare(g.dy(x,y)) +
				     dSquare(b.dx(x,y)) + dSquare(b.dy(x,y));
		// dx -   2      ,
		// iSquare(dx) -   4      ,
		//    3 ,    12
		grl.put(x,y, round(sqrt(isq)/12) );
	}
}
void matr3::GrLen ( matr3 &grl )					// grl :=    (a)
{
	this->test();
	//this->r.l2Grad( grl.r );
	//this->g.l2Grad( grl.g );
	//this->b.l2Grad( grl.b );
	grl.r.l2Grad(r);
	grl.g.l2Grad(g);
	grl.b.l2Grad(b);
}
void matr3::GrLenM( matr  &grl )					// grl :=     (a)
{
	this->test();
	this->test();
	int mx=this->Mx(), my=this->My();
	grl.setSize(mx,my);
	int x,y;
	for( y=0; y<my; y++ )
	for( x=0; x<mx; x++ )
	{
		//grl.put(x, y, round(this->mGrad(x, y)));
		grl.put(x, y, this->mGrad(x, y));
	}
}
/////////////////////////////////////////////////////////////////
double matr3::colorDif1(int x, int y, matr3 &f, double x1, double y1)//   this(x,y)  f(x1,y1) 
{
	double dr = f.r.getInterp1(x1, y1) - r(x, y);
	double dg = f.g.getInterp1(x1, y1) - g(x, y);
	double db = f.b.getInterp1(x1, y1) - b(x, y);
	return sqrt((dr*dr + dg*dg + db*db) / 3);
}
double matr3::colorDiff(int x, int y, matr3 &f, double x1, double y1)//   this(x,y)  f(x1,y1) 
{
	double dr = f.r.getInterp3(x1, y1) - r(x, y);
	double dg = f.g.getInterp3(x1, y1) - g(x, y);
	double db = f.b.getInterp3(x1, y1) - b(x, y);
	return sqrt((dr*dr + dg*dg + db*db) / 3);
}
/////////////////////////////////////////////////////////////////
double matr3::midVar()							//    
{
	double S = 0;
	int cnt = 0;
	for (int y = 0; y<My(); y++)
		for (int x = 0; x<Mx(); x++)
		{
			int r0 = r(x, y), g0=g(x,y), b0=b(x,y);
			if (y > 0) {
				S += iabs(r(x, y - 1) - r0) + iabs(g(x, y - 1) - b0) + iabs(b(x, y - 1) - b0);
				cnt++;
			}
			if (x > 0) {
				S += iabs(r(x - 1, y) - r0) + iabs(g(x - 1, y) - b0) + iabs(b(x - 1, y) - b0);
				cnt++;
			}
		}
	return S / (3*cnt);
}
/////////////////////////////////////////////////////////////////
void	matr3::transpose( matr3 &a)				// this := transpose(a)
{
	r.transpose(a.r);
	g.transpose(a.g);
	b.transpose(a.b);
}

//===========================================================================================
//   
//===========================================================================================

//-----------------------------------------------------------------------------
void toGray( matr &dst, matr3 &src )	// dst := grey(src)
{
	int mx=src.Mx(), my=src.My();
	dst.setSize(mx,my);
	int x,y;
	int r,g,b, iy;
	for( y=0; y<my; y++ )
	for( x=0; x<mx; x++ )
	{
		r = src.r.get(x,y);
		g = src.g.get(x,y);
		b = src.b.get(x,y);

		//iy = clip256((19595*r +38470*g + 7471*b)>>16);
		iy = (19595*r +38470*g + 7471*b)>>16;
		dst.put(x,y,iy);
	}
}
//-----------------------------------------------------------------------------
void mix( matr3 &res, matr3 &a, matr3 &b, ldouble al ) // res:= al*a + (1-al)*b
{
	int mx=a.Mx(), my=a.My();
	if( b.Mx() != mx || b.My() != my ) { res.setSize(0,0); return; }
	res.setSize(mx,my);
	ldouble be = 1-al;
	for( int y=0; y<my; y++ )
	for( int x=0; x<mx; x++ )
		res.put(x,y,
			round(al*a.r.get(x,y) + be*b.r.get(x,y) ),
			round(al*a.g.get(x,y) + be*b.g.get(x,y) ),
			round(al*a.b.get(x,y) + be*b.b.get(x,y) ) );
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------


