// iMatrix3.h:     matr
// 2006-11-07
//
#ifdef IMATRIX3
#error Secondary include
#endif
#define IMATRIX3
#ifndef IMATRIX
#error iMatrix.h not included
#endif
//  ,   .
//    -   
void    vError(const char *format, ...);		// show error function

class matr3{
public:
	matr r;
	matr g;
	matr b;

	// constructors/destructor
	matr3(int mx_=0, int my_=0);
	~matr3();
	// memory, sizes
	void	release();		// release memory
	void	clear();		//   
	void	setSize		(int mx_, int my_);	//  
	//    ,    ,  ,   
	void	setSizeClear(int mx_, int my_);	//     
	void	copyFrom(matr3 &a);
	void	swap(matr3 &a);					// this <-> a
	// tests
	void	test();
    void	testSameSize(matr  &a); 
    void	testSameSize(matr3 &a); 
	void	testIndexes (int x, int y);		// 0<=x,y<max
	void	normaIndexes(int &x, int &y);	//     0<=x,y<max
	// get/put
	int		Mx() { return r.Mx(); }
	int		My() { return r.My(); }
	int		getSize(){ return r.Mx()*r.My(); }
	void	put( int x, int y, int vr, int vg, int vb );
	int		get( int x, int y );							// r + 256*g + 256^2*b
	void	inc( int x, int y, int vr, int vg, int vb );	//     vr,vg,vb
	void	safePut( int x, int y, int vr, int vg, int vb );//  ,       
	void	set( int vr, int vg, int vb );					//     

	void	normalize( int k );					//       [0..k]
	void	Mod( int k );						//     mod k (0..k-1)

	// algebra
	void	add ( matr3 &a);					// this += a
	void	add ( matr3 &a, int k);				// this += a*k
	void	mult( int k);						// this *= k
	void	div ( int k);						// this /= k
	void	addV( int ar, int ag, int ab);		// this.r[*,*] += ar, this.g[*,*] += ag, this.b[*,*] += ab

	int     mGrad( int x, int y);				// .  : . .      
	ldouble	dist(int x1, int y1, int x2, int y2);//    

	// in/out
	void	saveText( const char *fname, const char *msg="", char *format = " (%4d,%4d=>%6d %6d %6d)" );
	void	copyBlock( matr3 &a, int xSrc, int ySrc, int xDst, int yDst, int dx, int dy );// this:=a[]

	void	smooth( int rad, int kind=1 );		//    rad
	void	decrease( matr3 &a, int k );		// this :=    k (a)
	void	resize( matr3 &a, int mx1, int my1);// this := resize(a)  (mx,my)
	void	diffX( matr3 &a );					// this := 2*da/dx 
	void	diffY( matr3 &a );					// this := 2*da/dy 
	ldouble  MSE( matr3 &a );					// Mean Square Error ( ) (this,a)
	ldouble  SSIM(matr3 &a);					// structural similarity (SSIM) index (this,a)  [-1..1]
	ldouble  DSSIM(matr3 &a);					// =(1-SSIM)/2									[ 1..0]

	void	toRGB  ( matr &rgb );				//      rgb
	void	fromRGB( matr &rgb );				//      rgb

	// bmp
	int		loadBMP    ( const char *fname );	
	int		saveBMP    ( const char *fname );
    void	drawEdge( int x0, int y0, int x1, int y1 ); //   [(x0,y0),(x1,y1)]  
    void	drawEdge( int x0, int y0, int x1, int y1, int cr, int cg, int cb ); //   [(x0,y0),(x1,y1)]  (cr,cb,cb)

	void GrLen3( matr  &grl );					// grl :=    (a)
	void GrLen ( matr3 &grl );					// grl :=    (a)
	void GrLenM( matr  &grl );					// grl :=     (a)
	double colorDif1(int x, int y, matr3 &f, double x1, double y1);//   this(x,y)  f(x1,y1) 
	double colorDiff(int x, int y, matr3 &f, double x1, double y1);//   this(x,y)  f(x1,y1) 

	double midVar();							//    


	void	transpose( matr3 &a);				// this := transpose(a)
};

//===========================================================================================
//   
//===========================================================================================

void toGray( matr &dst, matr3 &src );	// dst := grey(src)		ITU-601
void mix( matr3 &res, matr3 &a, matr3 &b, ldouble al ); // res:= al*a + (1-al)*b

