// vectU.cpp 2019-12-15   
// (c) Khashin S.I. khash2@gmail.com http://math.ivanovo.ac.ru/dalgebra/Khashin/index_r.html
//
//////////////////////////////////////////////////////////////////////
#ifdef _MSC_VER
#pragma warning(disable : 4996)
#pragma warning(disable : 4267)
#pragma warning(disable : 4018)
#endif

#ifdef __GNUC__
#pragma GCC diagnostic ignored "-Wwrite-strings"
#pragma GCC diagnostic ignored "-Wformat"
#endif

#include <stdio.h>
#include <vector>
#include <string>
#include <algorithm>
using namespace std;


#include "vUtil.h"			//   
#include "vectU.h"			//   


//----------------------------------------------------------------------------
void Vset(vector<double>&v, double x)// v[i]=x
{
	for (size_t i = 0; i < v.size(); i++) v[i] = x;
}
void VsetP(vector<double>&v, double*x)// v[i]=x[i]
{
	for (size_t i = 0; i < v.size(); i++) v[i] = x[i];
}
void Vset(vector<double>&v, double x0, double x1)// v[i]=random(x0,x1)
{
	for (size_t i = 0; i < v.size(); i++) v[i] = x0 + hRand()*(x1 - x0);
}
//  -------------------------------------------------------------------
void   Vadd(vector<double>&v, vector<double> &dv, double k)// v[i]+=x[i]*k
{
	for (size_t i = 0; i < v.size(); i++) v[i] += dv[i] * k;
}
void   Vadd(vector<double>&v, vector<double> &v0, vector<double> &dv, double k)// v=v0+k*dv
{
	for (size_t i = 0; i < v.size(); i++) v[i] = v0[i] + dv[i] * k;
}
void   VaddK(vector<double>&v, double k)// v[i]+=k
{
	for (size_t i = 0; i < v.size(); i++) v[i] += k;
}
//  ------------------------------------------------------------------
double Vnorma(vector<double>&v, double lam)	//     lam.    
{
	double len = 0;
	for (size_t i = 0; i < v.size(); i++) len += dSquare(v[i]);
	len = sqrt(len);
	if (len == 0) return 0;
	double co = lam / len;
	for (size_t i = 0; i < v.size(); i++) v[i] *= co;
	return len;
}
void Vmult(vector<double>&v, double x)		// v *= x
{
	for (size_t i = 0; i < v.size(); i++) v[i] *= x;
}
//  -------------------------------------------------------------------
double Vlen(vector<double>&v)					//  
{
	double len = 0;
	for (size_t i = 0; i < v.size(); i++) len += dSquare(v[i]);
	return sqrt(len);
}
double Vlen2(vector<double>&v)					//  ^2
{
	double len = 0;
	for (size_t i = 0; i < v.size(); i++) len += dSquare(v[i]);
	return len;
}
double Vmax(vector<double>&v)					// 
{
	double M = -dbl_MAX;
	for (size_t i = 0; i < v.size(); i++)
		if (v[i]>M)
			M = v[i];
	return M;
}
double Vmin(vector<double>&v)					// 
{
	double M = dbl_MAX;
	for (size_t i = 0; i < v.size(); i++)
		if (v[i]<M)
			M = v[i];
	return M;
}
double VmaxM(vector < double>&v)	//  
{
	double M = 0;
	for (size_t i = 0; i < v.size(); i++)
		if (fabs(v[i])>M)
			M = fabs(v[i]);
	return M;
}
double VminM(vector<double>&v)					//  
{
	double M = dbl_MAX;
	for (size_t i = 0; i < v.size(); i++)
		if (fabs(v[i])<M)
			M = fabs(v[i]);
	return M;
}

int Vargmax(vector<double>&v)				//  
{
	int ii = -1;
	double M = -dbl_MAX;
	for (size_t i = 0; i < v.size(); i++)
		if (v[i] > M) {
			M = v[i];
			ii = i;
		}
	return ii;
}
int Vargmax(vector<double>&v, double &w)		//  
{
	int ii = -1;
	w = -dbl_MAX;
	for (size_t i = 0; i < v.size(); i++)
		if (v[i] > w) {
			w = v[i];
			ii = i;
		}
	return ii;
}

int Vargmin(vector<double>&v)				//  
{
	int ii = -1;
	double M = dbl_MAX;
	for (size_t i = 0; i < v.size(); i++)
		if (v[i] < M) {
			M = v[i];
			ii = i;
		}
	return ii;
}
int VargmaxM(vector<double>&v)				//   
{
	int ii = -1;
	double M = -1;
	for (size_t i = 0; i < v.size(); i++)
		if (fabs(v[i]) > M) {
			M = fabs(v[i]);
			ii = i;
		}
	return ii;
}
int VargminM(vector<double>&v)				//   
{
	int ii = -1;
	double M = dbl_MAX;
	for (size_t i = 0; i < v.size(); i++)
		if (fabs(v[i]) < M) {
			M = fabs(v[i]);
			ii = i;
		}
	return ii;
}
double Vsum(vector<double>&v)					//  
{
	double s = 0;
	for (size_t i = 0; i < v.size(); i++) s += v[i];
	return s;
}
double VsumM(vector<double>&v)					//   
{
	double s = 0;
	for (size_t i = 0; i < v.size(); i++) s += fabs(v[i]);
	return s;
}
double Vdist(vector<double>&v, vector<double>&w)//   
{
	double s = 0;
	for (size_t i = 0; i < v.size(); i++) s += dSquare(v[i] - w[i]);
	return sqrt(s);
}
double Vprod(vector<double>&v, vector<double>&w)//  
{
	double s = 0;
	for (size_t i = 0; i < v.size(); i++) s += v[i]*w[i];
	return sqrt(s);
}
double Vcos(vector<double>&v, vector<double>&w)//    
{
	double sv = 0, sw = 0, svw = 0;
	for (size_t i = 0; i < v.size(); i++) {
		double vi = v[i], wi = w[i];
		sv += vi*vi;
		sw += wi*wi;
		svw += vi*wi;
	}
	sv = sqrt(sv*sw);
	if (sv > 1)sv = 1;
	if (sv < -1)sv = -1;
	return sv == 0 ? 0 : svw / sv;
}

bool   VinBox(vector<double>&min, vector<double>&v, vector<double>&max)	// min[i]<=v[i]<=max[i]
{
	for (size_t i = 0; i<v.size(); i++)
		if (min[i]>v[i] || v[i]>max[i]) return false;
	return true;
}

//-----------------------------------------------------------------------------
//         (v,w)
//  res =  ----- w
//         (w,w)
void   VProj(vector<double>&res, vector<double>&v, vector<double>&w)// res:=   v  w (res    v)
{
	int N = v.size();
	double vw = 0, ww = 0;
	for (int i = 0; i < N; i++) {
		ww += w[i] * w[i];
		vw += v[i] * w[i];
	}
	res.resize(N);
	if (ww == 0) {
		fill(res.begin(), res.end(), 0);
		return;
	}
	// ww != 0
	double co = vw / ww;
	for (int i = 0; i < N; i++) {
		res[i] = co * w[i];
	}
}
//-----------------------------------------------------------------------------
//              (v,w)
//  res = v  -  ----- w
//              (w,w)
void   VProjO(vector<double>&res, vector<double>&v, vector<double>&w)// res:= v - proj(v,w) (res    v)
{
	int N = v.size();
	double vw = 0, ww = 0;
	for (int i = 0; i < N; i++) {
		ww += w[i] * w[i];
		vw += v[i] * w[i];
	}
	res.resize(N);
	if (ww == 0) {
		res = v;
		return;
	}
	// ww != 0
	double co = vw / ww;
	for (int i = 0; i < N; i++) {
		res[i] = v[i] - co * w[i];
	}
}
//----------------------------------------------------------------------------
bool nextK(vector<int>&K, int d) // find next of K[i]
{
	int n = K.size();
	//  K   <d
	int i, sum;
	for (i = 0, sum = 0; i<n; i++)
	{
		sum += K[i];
		if (sum == d)	//  K[0]+..+K[i]=d
		{
			if (i == 0) return false;	//   
										// i>0  K[0]+..+K[i]=d. ,   K[j]=0
			K[i] = 0;
			K[i - 1]++;
			return true;
		}
	}
	//   <d
	K[n - 1]++;
	return true;
}
//----------------------------------------------------------------------------
double monom(vector<int>&K, vector<vector<double>> &a) // calculate one monom with powers K[i]
{
	double res = 1;
	for (int i = 0; i<K.size(); i++)
		res *= a[i][K[i]];			// res *= x[i]^K[i]
	return res;
}

//-----------------------------------------------------------------------------
// v :=     <=d    x.   v=binomial(n+d,d)
void plukker(vector<double>&x, int d, vector<double> &v)
{
	int n = x.size();
	vector<vector<double>> a; // a[i,j] = x[i]^j
	a.resize(n);
	for (int i = 0; i < n; i++) {
		a[i].clear();
		double xx = 1;
		a[i].push_back(xx);
		for (int j = 0; j < d; j++) {
			xx *= x[n - 1 - i];
			a[i].push_back(xx);
		}
	}
	v.clear();
	vector<int>K(n);
	do {
		v.push_back(monom(K, a));	//   
	} while (nextK(K, d));
}
//----------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------
string VtoString(vector<double>&v, const char *format)	//   - 
{
	string s = "";
	char buf[64];
	for (size_t i = 0; i < v.size(); i++) {
		sprintf(buf, format, v[i]);
		s += buf;
	}
	return s;
}
//-----------------------------------------------------------------------------
void   VsaveA(FILE *f, vector<double>&v, const char *msg, const char *format)//      f
{
	fprintf(f, "%-8s, ", msg);
	for (size_t i = 0; i < v.size(); i++)
		fprintf(f, format, v[i]);
	fprintf(f, "\n");
}

void   VsaveS(const char* fname, vector<double>& v, const char* msg, const char* format)
{
	FILE* f = fopen(fname, "a");
	VsaveA(f, v, msg, format);
	fclose(f);
}

//-----------------------------------------------------------------------------
string HreadString(FILE *f)			//   s    . False,    
{
	string s;
	char ch = fgetc(f);
	while ( !feof(f) && ch != '\n' && ch != '\r' ) {
		s.push_back(ch);
		ch = fgetc(f);
	}
	if (feof(f)) return s;
	ch = fgetc(f);
	if (ch != '\r') ungetc(ch, f);
	return s;
}
//-----------------------------------------------------------------------------
void   HreadFile(char *fname, vector<string> &lines)	//     
{
	FILE *f = fopen(fname, "r");
	if (!f) return;
	lines.clear();
	do {
		lines.push_back(HreadString(f));
	} while (!feof(f));
	fclose(f);
}
//-----------------------------------------------------------------------------
// ,  
// 1)     
// 2)    ','  ';'  '\0',  s =    
// 3)  .  : ",; \t\0",  s
// 4)    .    
char *HHgetItem(char *buf, string &s)
{
	char ch;
	while ((ch = *buf) == ' ' || ch == '\t') buf++;	//     
	s.clear();
	ch = *buf++;
	if (ch == 0) return buf-1;						//   0
	if (ch == ',' || ch == ';') return buf;			// s = <empty>
	// ch -   
	do {
		s.push_back(ch);
		ch = *buf++;
	} while (ch != ' ' && ch != '\t' && ch != ',' && ch != ';' && ch != 0);
	if (ch == 0) return buf - 1;					//   0
	if (ch == ',' || ch == ';') return buf;			//   " "
	while ((ch = *buf) == ' ' || ch == '\t') buf++;	//     
	return ++buf;
}
//-----------------------------------------------------------------------------
void   HparseC(char *buf, vector<string> &items)	//   buf  
{
	items.clear();
	string s;
	while (*buf) {
		buf = HHgetItem(buf, s);
		items.push_back(s);
	}
}
void   HparseS(string bf, vector<string> &items)		//   buf  
{
	HparseC((char*)bf.c_str(), items);
}
//-----------------------------------------------------------------------------
int Sindex(vector<string>&lines, string s)				//     
{
	for (size_t i = 0; i < lines.size(); i++)
		if (lines[i] == s) return i;
	return -1;
}
//-----------------------------------------------------------------------------
string SLtrim(string&s)								//   
{
	size_t first = s.find_first_not_of(" \t");
	return s.substr(first, 999999999);
}
string SRtrim(string&s)								//   
{
	size_t last = s.find_last_not_of(" \t");
	return s.substr(0, last + 1);
}
string Strim(string&s)								//     
{
	if (s.length() == 0) return s;
	size_t first = s.find_first_not_of(" \t");
	if (first == -1) return "";
	size_t last = s.find_last_not_of(" \t");
	return s.substr(first, (last - first + 1));
}
void   SSplit(string &s, string delim, string &left, string &right)//     delim     
{
	size_t first = s.find(delim);
	//left = Strim(s.substr(0, first));
	string st = s.substr(0, first);
	left = Strim(st);
	if (first == string::npos) right = "";
	else {
		string s1 = s.substr(first + delim.length());
		right = Strim(s1);
	}
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
